from __future__ import annotations

from collections import deque

import pyqtgraph as pg
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from pydantic import Field, ValidationError, field_validator
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget

from bec_widgets.utils import Colors, ConnectionConfig
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.side_panel import SidePanel
from bec_widgets.widgets.plots.multi_waveform.settings.control_panel import (
    MultiWaveformControlPanel,
)
from bec_widgets.widgets.plots.multi_waveform.toolbar_bundles.monitor_selection import (
    MultiWaveformSelectionToolbarBundle,
)
from bec_widgets.widgets.plots.plot_base import PlotBase

logger = bec_logger.logger


class MultiWaveformConfig(ConnectionConfig):
    color_palette: str | None = Field(
        "plasma", description="The color palette of the figure widget.", validate_default=True
    )
    curve_limit: int | None = Field(
        200, description="The maximum number of curves to display on the plot."
    )
    flush_buffer: bool | None = Field(
        False, description="Flush the buffer of the plot widget when the curve limit is reached."
    )
    monitor: str | None = Field(None, description="The monitor to set for the plot widget.")
    curve_width: int | None = Field(1, description="The width of the curve on the plot.")
    opacity: int | None = Field(50, description="The opacity of the curve on the plot.")
    highlight_last_curve: bool | None = Field(
        True, description="Highlight the last curve on the plot."
    )

    model_config: dict = {"validate_assignment": True}
    _validate_color_map_z = field_validator("color_palette")(Colors.validate_color_map)


class MultiWaveform(PlotBase):
    """
    MultiWaveform widget for displaying multiple waveforms emitted by a single signal.
    """

    PLUGIN = True
    RPC = True
    ICON_NAME = "ssid_chart"
    USER_ACCESS = [
        # General PlotBase Settings
        "enable_toolbar",
        "enable_toolbar.setter",
        "enable_side_panel",
        "enable_side_panel.setter",
        "enable_fps_monitor",
        "enable_fps_monitor.setter",
        "set",
        "title",
        "title.setter",
        "x_label",
        "x_label.setter",
        "y_label",
        "y_label.setter",
        "x_limits",
        "x_limits.setter",
        "y_limits",
        "y_limits.setter",
        "x_grid",
        "x_grid.setter",
        "y_grid",
        "y_grid.setter",
        "inner_axes",
        "inner_axes.setter",
        "outer_axes",
        "outer_axes.setter",
        "lock_aspect_ratio",
        "lock_aspect_ratio.setter",
        "auto_range_x",
        "auto_range_x.setter",
        "auto_range_y",
        "auto_range_y.setter",
        "x_log",
        "x_log.setter",
        "y_log",
        "y_log.setter",
        "legend_label_size",
        "legend_label_size.setter",
        # MultiWaveform Specific RPC Access
        "highlighted_index",
        "highlighted_index.setter",
        "highlight_last_curve",
        "highlight_last_curve.setter",
        "color_palette",
        "color_palette.setter",
        "opacity",
        "opacity.setter",
        "flush_buffer",
        "flush_buffer.setter",
        "max_trace",
        "max_trace.setter",
        "monitor",
        "monitor.setter",
        "set_curve_limit",
        "plot",
        "set_curve_highlight",
        "clear_curves",
    ]

    monitor_signal_updated = Signal()
    highlighted_curve_index_changed = Signal(int)

    def __init__(
        self,
        parent: QWidget | None = None,
        config: MultiWaveformConfig | None = None,
        client=None,
        gui_id: str | None = None,
        popups: bool = True,
        **kwargs,
    ):
        if config is None:
            config = MultiWaveformConfig(widget_class=self.__class__.__name__)
        super().__init__(
            parent=parent, config=config, client=client, gui_id=gui_id, popups=popups, **kwargs
        )

        # Scan Data
        self.old_scan_id = None
        self.scan_id = None
        self.connected = False
        self._current_highlight_index = 0
        self._curves = deque()
        self.visible_curves = []
        self.number_of_visible_curves = 0

        self._init_control_panel()

    ################################################################################
    # Widget Specific GUI interactions
    ################################################################################
    def _init_toolbar(self):
        self.monitor_selection_bundle = MultiWaveformSelectionToolbarBundle(
            bundle_id="motor_selection", target_widget=self
        )
        self.toolbar.add_bundle(self.monitor_selection_bundle, target_widget=self)
        super()._init_toolbar()
        self.toolbar.widgets["reset_legend"].action.setVisible(False)

    def _init_control_panel(self):
        self.control_panel = SidePanel(self, orientation="top", panel_max_width=90)
        self.layout_manager.add_widget_relative(
            self.control_panel, self.round_plot_widget, "bottom"
        )
        self.controls = MultiWaveformControlPanel(parent=self, target_widget=self)
        self.control_panel.add_menu(
            action_id="control",
            icon_name="tune",
            tooltip="Show Control panel",
            widget=self.controls,
            title=None,
        )
        self.control_panel.toolbar.widgets["control"].action.trigger()

    ################################################################################
    # Widget Specific Properties
    ################################################################################

    @property
    def curves(self) -> deque:
        """
        Get the curves of the plot widget as a deque.
        Returns:
            deque: Deque of curves.
        """
        return self._curves

    @curves.setter
    def curves(self, value: deque):
        self._curves = value

    @SafeProperty(int, designable=False)
    def highlighted_index(self):
        return self._current_highlight_index

    @highlighted_index.setter
    def highlighted_index(self, value: int):
        self._current_highlight_index = value
        self.property_changed.emit("highlighted_index", value)
        self.set_curve_highlight(value)

    @SafeProperty(bool)
    def highlight_last_curve(self) -> bool:
        """
        Get the highlight_last_curve property.
        Returns:
            bool: The highlight_last_curve property.
        """
        return self.config.highlight_last_curve

    @highlight_last_curve.setter
    def highlight_last_curve(self, value: bool):
        self.config.highlight_last_curve = value
        self.property_changed.emit("highlight_last_curve", value)
        self.set_curve_highlight(-1)

    @SafeProperty(str)
    def color_palette(self) -> str:
        """
        The color palette of the figure widget.
        """
        return self.config.color_palette

    @color_palette.setter
    def color_palette(self, value: str):
        """
        Set the color palette of the figure widget.

        Args:
            value(str): The color palette to set.
        """
        try:
            self.config.color_palette = value
        except ValidationError:
            return
        self.set_curve_highlight(self._current_highlight_index)
        self._sync_monitor_selection_toolbar()

    @SafeProperty(int)
    def opacity(self) -> int:
        """
        The opacity of the figure widget.
        """
        return self.config.opacity

    @opacity.setter
    def opacity(self, value: int):
        """
        Set the opacity of the figure widget.

        Args:
            value(int): The opacity to set.
        """
        self.config.opacity = max(0, min(100, value))
        self.property_changed.emit("opacity", value)
        self.set_curve_highlight(self._current_highlight_index)

    @SafeProperty(bool)
    def flush_buffer(self) -> bool:
        """
        The flush_buffer property.
        """
        return self.config.flush_buffer

    @flush_buffer.setter
    def flush_buffer(self, value: bool):
        self.config.flush_buffer = value
        self.property_changed.emit("flush_buffer", value)
        self.set_curve_limit(
            max_trace=self.config.curve_limit, flush_buffer=self.config.flush_buffer
        )

    @SafeProperty(int)
    def max_trace(self) -> int:
        """
        The maximum number of traces to display on the plot.
        """
        return self.config.curve_limit

    @max_trace.setter
    def max_trace(self, value: int):
        """
        Set the maximum number of traces to display on the plot.

        Args:
            value(int): The maximum number of traces to display.
        """
        self.config.curve_limit = value
        self.property_changed.emit("max_trace", value)
        self.set_curve_limit(
            max_trace=self.config.curve_limit, flush_buffer=self.config.flush_buffer
        )

    @SafeProperty(str)
    def monitor(self) -> str:
        """
        The monitor of the figure widget.
        """
        return self.config.monitor

    @monitor.setter
    def monitor(self, value: str):
        """
        Set the monitor of the figure widget.

        Args:
            value(str): The monitor to set.
        """
        self.plot(value)

    ################################################################################
    # High Level methods for API
    ################################################################################

    @SafeSlot(popup_error=True)
    def plot(self, monitor: str, color_palette: str | None = "plasma"):
        """
        Create a plot for the given monitor.
        Args:
            monitor (str): The monitor to set.
            color_palette (str|None): The color palette to use for the plot.
        """
        self.entry_validator.validate_monitor(monitor)
        self._disconnect_monitor()
        self.config.monitor = monitor
        self._connect_monitor()
        if color_palette is not None:
            self.color_palette = color_palette
        self._sync_monitor_selection_toolbar()

    @SafeSlot(int, bool)
    def set_curve_limit(self, max_trace: int, flush_buffer: bool):
        """
        Set the maximum number of traces to display on the plot.

        Args:
            max_trace (int): The maximum number of traces to display.
            flush_buffer (bool): Flush the buffer.
        """
        if max_trace != self.config.curve_limit:
            self.config.curve_limit = max_trace
        if flush_buffer != self.config.flush_buffer:
            self.config.flush_buffer = flush_buffer

        if self.config.curve_limit is None:
            self.scale_colors()
            return

        if self.config.flush_buffer:
            # Remove excess curves from the plot and the deque
            while len(self.curves) > self.config.curve_limit:
                curve = self.curves.popleft()
                self.plot_item.removeItem(curve)
        else:
            # Hide or show curves based on the new max_trace
            num_curves_to_show = min(self.config.curve_limit, len(self.curves))
            for i, curve in enumerate(self.curves):
                if i < len(self.curves) - num_curves_to_show:
                    curve.hide()
                else:
                    curve.show()
        self.scale_colors()
        self.monitor_signal_updated.emit()

    ################################################################################
    # BEC Update Methods
    ################################################################################
    @SafeSlot(dict, dict)
    def on_monitor_1d_update(self, msg: dict, metadata: dict):
        """
        Update the plot widget with the monitor data.

        Args:
            msg(dict): The message data.
            metadata(dict): The metadata of the message.
        """
        data = msg.get("data", None)
        current_scan_id = metadata.get("scan_id", None)

        if current_scan_id != self.scan_id:
            self.scan_id = current_scan_id
            self.clear_curves()
            self.curves.clear()
            if self.crosshair:
                self.crosshair.clear_markers()

        # Always create a new curve and add it
        curve = pg.PlotDataItem()
        curve.setData(data)
        self.plot_item.addItem(curve)
        self.curves.append(curve)

        # Max Trace and scale colors
        self.set_curve_limit(self.config.curve_limit, self.config.flush_buffer)

    @SafeSlot(int)
    def set_curve_highlight(self, index: int):
        """
        Set the curve highlight based on visible curves.

        Args:
            index (int): The index of the curve to highlight among visible curves.
        """
        self.plot_item.visible_curves = [curve for curve in self.curves if curve.isVisible()]
        num_visible_curves = len(self.plot_item.visible_curves)
        self.number_of_visible_curves = num_visible_curves

        if num_visible_curves == 0:
            return  # No curves to highlight

        if index >= num_visible_curves:
            index = num_visible_curves - 1
        elif index < 0:
            index = num_visible_curves + index
        self._current_highlight_index = index
        num_colors = num_visible_curves
        colors = Colors.evenly_spaced_colors(
            colormap=self.config.color_palette, num=num_colors, format="HEX"
        )
        for i, curve in enumerate(self.plot_item.visible_curves):
            curve.setPen()
            if i == self._current_highlight_index:
                curve.setPen(pg.mkPen(color=colors[i], width=5))
                curve.setAlpha(alpha=1, auto=False)
                curve.setZValue(1)
            else:
                curve.setPen(pg.mkPen(color=colors[i], width=1))
                curve.setAlpha(alpha=self.config.opacity / 100, auto=False)
                curve.setZValue(0)

        self.highlighted_curve_index_changed.emit(self._current_highlight_index)

    def _disconnect_monitor(self):
        try:
            previous_monitor = self.config.monitor
        except AttributeError:
            previous_monitor = None

        if previous_monitor and self.connected is True:
            self.bec_dispatcher.disconnect_slot(
                self.on_monitor_1d_update, MessageEndpoints.device_monitor_1d(previous_monitor)
            )
            self.connected = False

    def _connect_monitor(self):
        """
        Connect the monitor to the plot widget.
        """

        if self.config.monitor and self.connected is False:
            self.bec_dispatcher.connect_slot(
                self.on_monitor_1d_update, MessageEndpoints.device_monitor_1d(self.config.monitor)
            )
            self.connected = True

    ################################################################################
    # Utility Methods
    ################################################################################
    def scale_colors(self):
        """
        Scale the colors of the curves based on the current colormap.
        """
        # TODO probably has to be changed to property
        if self.config.highlight_last_curve:
            self.set_curve_highlight(-1)  # Use -1 to highlight the last visible curve
        else:
            self.set_curve_highlight(self._current_highlight_index)

    def hook_crosshair(self) -> None:
        """
        Specific hookfor crosshair, since it is for multiple curves.
        """
        super().hook_crosshair()
        if self.crosshair:
            self.highlighted_curve_index_changed.connect(self.crosshair.update_highlighted_curve)
            if self.curves:
                self.crosshair.update_highlighted_curve(self._current_highlight_index)

    def clear_curves(self):
        """
        Remove all curves from the plot, excluding crosshair items.
        """
        items_to_remove = []
        for item in self.plot_item.items:
            if not getattr(item, "is_crosshair", False) and isinstance(item, pg.PlotDataItem):
                items_to_remove.append(item)
        for item in items_to_remove:
            self.plot_item.removeItem(item)

    def _sync_monitor_selection_toolbar(self):
        """
        Sync the motor map selection toolbar with the current motor map.
        """
        if self.monitor_selection_bundle is not None:
            monitor = self.monitor_selection_bundle.monitor.currentText()
            color_palette = self.monitor_selection_bundle.colormap_widget.colormap

            if monitor != self.config.monitor:
                self.monitor_selection_bundle.monitor.blockSignals(True)
                self.monitor_selection_bundle.monitor.set_device(self.config.monitor)
                self.monitor_selection_bundle.monitor.check_validity(self.config.monitor)
                self.monitor_selection_bundle.monitor.blockSignals(False)

            if color_palette != self.config.color_palette:
                self.monitor_selection_bundle.colormap_widget.blockSignals(True)
                self.monitor_selection_bundle.colormap_widget.colormap = self.config.color_palette
                self.monitor_selection_bundle.colormap_widget.blockSignals(False)

    def cleanup(self):
        self._disconnect_monitor()
        self.clear_curves()
        self.monitor_selection_bundle.cleanup()
        super().cleanup()



"""This custom class is a thin wrapper around the SignalProxy class to allow signal calls to be blocked.
Unblocking the proxy needs to be done through the slot unblock_proxy. The most likely use case for this class is
when the callback function is potentially initiating a slower progress, i.e. requesting a data analysis routine to
analyse data. Requesting a new fit may lead to request piling up and an overall slow done of performance. This proxy
will allow you to decide by yourself when to unblock and execute the callback again."""

from pyqtgraph import SignalProxy
from qtpy.QtCore import QTimer, Signal

from bec_widgets.utils.error_popups import SafeSlot


class BECSignalProxy(SignalProxy):
    """
    Thin wrapper around the SignalProxy class to allow signal calls to be blocked,
    but arguments still being stored.

    Args:
        *args: Arguments to pass to the SignalProxy class.
        rateLimit (int): The rateLimit of the proxy.
        timeout (float): The number of seconds after which the proxy automatically
                         unblocks if still blocked. Default is 10.0 seconds.
        **kwargs: Keyword arguments to pass to the SignalProxy class.

    Example:
        >>> proxy = BECSignalProxy(signal, rate_limit=25, slot=callback)
    """

    is_blocked = Signal(bool)

    def __init__(self, *args, rateLimit=25, timeout=10.0, **kwargs):
        super().__init__(*args, rateLimit=rateLimit, **kwargs)
        self._blocking = False
        self.old_args = None
        self.new_args = None

        # Store timeout value (in seconds)
        self._timeout = timeout

        # Create a single-shot timer for auto-unblocking
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._timeout_unblock)

    @property
    def blocked(self):
        """Returns if the proxy is blocked"""
        return self._blocking

    @blocked.setter
    def blocked(self, value: bool):
        self._blocking = value
        self.is_blocked.emit(value)

    def signalReceived(self, *args):
        """Receive signal, store the args and call signalReceived from the parent class if not blocked"""
        self.new_args = args
        if self.blocked is True:
            return
        self.blocked = True
        self.old_args = args
        super().signalReceived(*args)

        self._timer.start(int(self._timeout * 1000))

    @SafeSlot()
    def unblock_proxy(self):
        """Unblock the proxy, and call the signalReceived method in case there was an update of the args."""
        if self.blocked:
            self._timer.stop()
            self.blocked = False
            if self.new_args != self.old_args:
                self.signalReceived(*self.new_args)

    @SafeSlot()
    def _timeout_unblock(self):
        """
        Internal method called by the QTimer upon timeout. Unblocks the proxy
        automatically if it is still blocked.
        """
        if self.blocked:
            self.unblock_proxy()

    def cleanup(self):
        """
        Cleanup the proxy by stopping the timer and disconnecting the timeout signal.
        """
        self._timer.stop()
        self._timer.timeout.disconnect(self._timeout_unblock)
        self._timer.deleteLater()



from __future__ import annotations

from typing import TYPE_CHECKING

import darkdetect
from bec_lib.logger import bec_logger
from qtpy.QtCore import QObject, Slot
from qtpy.QtWidgets import QApplication

from bec_widgets.cli.rpc.rpc_register import RPCRegister
from bec_widgets.utils.bec_connector import BECConnector, ConnectionConfig
from bec_widgets.utils.colors import set_theme

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.widgets.containers.dock import BECDock

logger = bec_logger.logger


class BECWidget(BECConnector):
    """Mixin class for all BEC widgets, to handle cleanup"""

    # The icon name is the name of the icon in the icon theme, typically a name taken
    # from fonts.google.com/icons. Override this in subclasses to set the icon name.
    ICON_NAME = "widgets"
    USER_ACCESS = ["remove"]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        client=None,
        config: ConnectionConfig = None,
        gui_id: str | None = None,
        theme_update: bool = False,
        parent_dock: BECDock | None = None,  # TODO should go away -> issue created #473
        **kwargs,
    ):
        """
        Base class for all BEC widgets. This class should be used as a mixin class for all BEC widgets, e.g.:


        >>> class MyWidget(BECWidget, QWidget):
        >>>     def __init__(self, parent=None, client=None, config=None, gui_id=None):
        >>>         super().__init__(client=client, config=config, gui_id=gui_id)
        >>>         QWidget.__init__(self, parent=parent)


        Args:
            client(BECClient, optional): The BEC client.
            config(ConnectionConfig, optional): The connection configuration.
            gui_id(str, optional): The GUI ID.
            theme_update(bool, optional): Whether to subscribe to theme updates. Defaults to False. When set to True, the
                widget's apply_theme method will be called when the theme changes.
        """

        super().__init__(
            client=client, config=config, gui_id=gui_id, parent_dock=parent_dock, **kwargs
        )
        if not isinstance(self, QObject):
            raise RuntimeError(f"{repr(self)} is not a subclass of QWidget")
        app = QApplication.instance()
        if not hasattr(app, "theme"):
            # DO NOT SET THE THEME TO AUTO! Otherwise, the qwebengineview will segfault
            # Instead, we will set the theme to the system setting on startup
            if darkdetect.isDark():
                set_theme("dark")
            else:
                set_theme("light")

        if theme_update:
            logger.debug(f"Subscribing to theme updates for {self.__class__.__name__}")
            self._connect_to_theme_change()

    def _connect_to_theme_change(self):
        """Connect to the theme change signal."""
        qapp = QApplication.instance()
        if hasattr(qapp, "theme_signal"):
            qapp.theme_signal.theme_updated.connect(self._update_theme)

    def _update_theme(self, theme: str | None = None):
        """Update the theme."""
        if theme is None:
            qapp = QApplication.instance()
            if hasattr(qapp, "theme"):
                theme = qapp.theme.theme
            else:
                theme = "dark"
        self.apply_theme(theme)

    @Slot(str)
    def apply_theme(self, theme: str):
        """
        Apply the theme to the widget.

        Args:
            theme(str, optional): The theme to be applied.
        """

    def cleanup(self):
        """Cleanup the widget."""
        with RPCRegister.delayed_broadcast():
            # All widgets need to call super().cleanup() in their cleanup method
            logger.info(f"Registry cleanup for widget {self.__class__.__name__}")
            self.rpc_register.remove_rpc(self)

    def closeEvent(self, event):
        """Wrap the close even to ensure the rpc_register is cleaned up."""
        try:
            if not self._destroyed:
                self.cleanup()
                self._destroyed = True
        finally:
            super().closeEvent(event)  # pylint: disable=no-member



from __future__ import annotations

from bec_widgets.widgets.containers.auto_update.auto_updates import AutoUpdates
from bec_widgets.widgets.containers.dock.dock_area import BECDockArea


def dock_area(object_name: str | None = None) -> BECDockArea:
    _dock_area = BECDockArea(object_name=object_name, root_widget=True)
    return _dock_area


def auto_update_dock_area(object_name: str | None = None) -> AutoUpdates:
    """
    Create a dock area with auto update enabled.

    Args:
        object_name(str): The name of the dock area.

    Returns:
        BECDockArea: The created dock area.
    """
    _auto_update = AutoUpdates(object_name=object_name)
    return _auto_update



from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.toolbar import MaterialIconAction, ToolbarBundle


class SaveStateBundle(ToolbarBundle):
    """
    A bundle of actions that are hooked in this constructor itself,
    so that you can immediately connect the signals and toggle states.

    This bundle is for a toolbar that controls saving the state of the widget.
    """

    def __init__(self, bundle_id="mouse_interaction", target_widget=None, **kwargs):
        super().__init__(bundle_id=bundle_id, actions=[], **kwargs)
        self.target_widget = target_widget

        # Create each MaterialIconAction with a parent
        # so the signals can fire even if the toolbar isn't added yet.
        save_state = MaterialIconAction(
            icon_name="download", tooltip="Save Widget State", parent=self.target_widget
        )
        load_state = MaterialIconAction(
            icon_name="upload", tooltip="Load Widget State", parent=self.target_widget
        )

        # Add them to the bundle
        self.add_action("save", save_state)
        self.add_action("matplotlib", load_state)

        # Immediately connect signals
        save_state.action.triggered.connect(self.save_state_dialog)
        load_state.action.triggered.connect(self.load_state_dialog)

    @SafeSlot()
    def save_state_dialog(self):
        """
        Open the export dialog to save a state of the widget.
        """
        if self.target_widget:
            self.target_widget.state_manager.save_state()

    @SafeSlot()
    def load_state_dialog(self):
        """
        Load a saved state of the widget.
        """
        if self.target_widget:
            self.target_widget.state_manager.load_state()



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from bec_lib.logger import bec_logger
from bec_qthemes._icon.material_icons import material_icon
from qtpy.QtGui import QColor
from qtpy.QtWidgets import (
    QColorDialog,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpinBox,
    QToolButton,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils import ConnectionConfig, EntryValidator
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import Colors
from bec_widgets.utils.toolbar import MaterialIconAction, ModularToolBar
from bec_widgets.widgets.control.device_input.device_line_edit.device_line_edit import (
    DeviceLineEdit,
)
from bec_widgets.widgets.dap.dap_combo_box.dap_combo_box import DapComboBox
from bec_widgets.widgets.plots.waveform.curve import CurveConfig, DeviceSignal
from bec_widgets.widgets.utility.visual.color_button_native.color_button_native import (
    ColorButtonNative,
)
from bec_widgets.widgets.utility.visual.colormap_widget.colormap_widget import BECColorMapWidget

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.widgets.plots.waveform.waveform import Waveform


logger = bec_logger.logger


class CurveRow(QTreeWidgetItem):
    DELETE_BUTTON_COLOR = "#CC181E"
    """A unified row that can represent either a device or a DAP curve.

    Columns:
      0: Actions (delete or "Add DAP" if source=device)
      1..2: DeviceLineEdit and QLineEdit if source=device, or "Model" label and DapComboBox if source=dap
      3: ColorButton
      4: Style QComboBox
      5: Pen width QSpinBox
      6: Symbol size QSpinBox
    """

    def __init__(
        self,
        tree: QTreeWidget,
        parent_item: QTreeWidgetItem | None = None,
        config: CurveConfig | None = None,
        device_manager=None,
    ):
        if parent_item:
            super().__init__(parent_item)
        else:
            # A top-level device row.
            super().__init__(tree)

        self.tree = tree
        self.parent_item = parent_item
        self.curve_tree = tree.parent()  # The CurveTree widget
        self.curve_tree.all_items.append(self)  # Track stable ordering

        # BEC user input
        self.device_edit = None
        self.dap_combo = None

        self.dev = device_manager
        self.entry_validator = EntryValidator(self.dev)

        self.config = config or CurveConfig()
        self.source = self.config.source
        self.dap_rows = []

        # Create column 0 (Actions)
        self._init_actions()
        # Create columns 1..2, depending on source
        self._init_source_ui()
        # Create columns 3..6 (color, style, width, symbol)
        self._init_style_controls()

    def _init_actions(self):
        """Create the actions widget in column 0, including a delete button and maybe 'Add DAP'."""
        self.actions_widget = QWidget()
        actions_layout = QHBoxLayout(self.actions_widget)
        actions_layout.setContentsMargins(0, 0, 0, 0)
        actions_layout.setSpacing(0)

        # Delete button
        self.delete_button = QToolButton()
        delete_icon = material_icon(
            "delete",
            size=(20, 20),
            convert_to_pixmap=False,
            filled=False,
            color=self.DELETE_BUTTON_COLOR,
        )
        self.delete_button.setIcon(delete_icon)
        self.delete_button.clicked.connect(lambda: self.remove_self())
        actions_layout.addWidget(self.delete_button)

        # If device row, add "Add DAP" button
        if self.source == "device":
            self.add_dap_button = QPushButton("DAP")
            self.add_dap_button.clicked.connect(lambda: self.add_dap_row())
            actions_layout.addWidget(self.add_dap_button)

        self.tree.setItemWidget(self, 0, self.actions_widget)

    def _init_source_ui(self):
        """Create columns 1 and 2. For device rows, we have device/entry edits; for dap rows, label/model combo."""
        if self.source == "device":
            # Device row: columns 1..2 are device line edits
            self.device_edit = DeviceLineEdit(parent=self.tree)
            self.entry_edit = QLineEdit(parent=self.tree)  # TODO in future will be signal line edit
            if self.config.signal:
                self.device_edit.setText(self.config.signal.name or "")
                self.entry_edit.setText(self.config.signal.entry or "")

            self.tree.setItemWidget(self, 1, self.device_edit)
            self.tree.setItemWidget(self, 2, self.entry_edit)

        else:
            # DAP row: column1= "Model" label, column2= DapComboBox
            self.label_widget = QLabel("Model")
            self.tree.setItemWidget(self, 1, self.label_widget)
            self.dap_combo = DapComboBox(parent=self.tree)
            self.dap_combo.populate_fit_model_combobox()
            # If config.signal has a dap
            if self.config.signal and self.config.signal.dap:
                dap_value = self.config.signal.dap
                idx = self.dap_combo.fit_model_combobox.findText(dap_value)
                if idx >= 0:
                    self.dap_combo.fit_model_combobox.setCurrentIndex(idx)
            else:
                self.dap_combo.select_fit_model("GaussianModel")  # default

            self.tree.setItemWidget(self, 2, self.dap_combo)

    def _init_style_controls(self):
        """Create columns 3..6: color button, style combo, width spin, symbol spin."""
        # Color in col 3
        self.color_button = ColorButtonNative(color=self.config.color)
        self.color_button.clicked.connect(lambda: self._select_color(self.color_button))
        self.tree.setItemWidget(self, 3, self.color_button)

        # Style in col 4
        self.style_combo = QComboBox()
        self.style_combo.addItems(["solid", "dash", "dot", "dashdot"])
        idx = self.style_combo.findText(self.config.pen_style)
        if idx >= 0:
            self.style_combo.setCurrentIndex(idx)
        self.tree.setItemWidget(self, 4, self.style_combo)

        # Pen width in col 5
        self.width_spin = QSpinBox()
        self.width_spin.setRange(1, 20)
        self.width_spin.setValue(self.config.pen_width)
        self.tree.setItemWidget(self, 5, self.width_spin)

        # Symbol size in col 6
        self.symbol_spin = QSpinBox()
        self.symbol_spin.setRange(1, 20)
        self.symbol_spin.setValue(self.config.symbol_size)
        self.tree.setItemWidget(self, 6, self.symbol_spin)

    def _select_color(self, button):
        """
        Selects a new color using a color dialog and applies it to the specified button. Updates
        related configuration properties based on the chosen color.

        Args:
            button: The button widget whose color is being modified.
        """
        current_color = QColor(button.color())
        chosen_color = QColorDialog.getColor(current_color, self.tree, "Select Curve Color")
        if chosen_color.isValid():
            button.set_color(chosen_color)
            self.config.color = chosen_color.name()
            self.config.symbol_color = chosen_color.name()

    def add_dap_row(self):
        """Create a new DAP row as a child. Only valid if source='device'."""
        if self.source != "device":
            return
        curve_tree = self.tree.parent()
        parent_label = self.config.label

        # Inherit device name/entry
        dev_name = ""
        dev_entry = ""
        if self.config.signal:
            dev_name = self.config.signal.name
            dev_entry = self.config.signal.entry

        # Create a new config for the DAP row
        dap_cfg = CurveConfig(
            widget_class="Curve",
            source="dap",
            parent_label=parent_label,
            signal=DeviceSignal(name=dev_name, entry=dev_entry),
        )
        new_dap = CurveRow(self.tree, parent_item=self, config=dap_cfg, device_manager=self.dev)
        # Expand device row to show new child
        self.tree.expandItem(self)

        # Give the new row a color from the buffer:
        curve_tree._ensure_color_buffer_size()
        idx = len(curve_tree.all_items) - 1
        new_col = curve_tree.color_buffer[idx]
        new_dap.color_button.set_color(new_col)
        new_dap.config.color = new_col
        new_dap.config.symbol_color = new_col

    def remove_self(self):
        """Remove this row from the tree and from the parent's item list."""
        # Recursively remove all child rows first
        for i in reversed(range(self.childCount())):
            child = self.child(i)
            if isinstance(child, CurveRow):
                child.remove_self()

        # Clean up the widget references if they still exist
        if getattr(self, "device_edit", None) is not None:
            self.device_edit.close()
            self.device_edit.deleteLater()
            self.device_edit = None

        if getattr(self, "dap_combo", None) is not None:
            self.dap_combo.close()
            self.dap_combo.deleteLater()
            self.dap_combo = None

        if getattr(self, "color_button", None) is not None:
            self.color_button.close()
            self.color_button.deleteLater()
            self.color_button = None

        # Remove the item from the tree widget
        index = self.tree.indexOfTopLevelItem(self)
        if index != -1:
            self.tree.takeTopLevelItem(index)
        elif self.parent_item:
            self.parent_item.removeChild(self)

        # Finally, remove self from the registration list in the curve tree
        curve_tree = self.tree.parent()
        if self in curve_tree.all_items:
            curve_tree.all_items.remove(self)

    def export_data(self) -> dict:
        """Collect data from the GUI widgets, update config, and return as a dict.

        Returns:
            dict: The serialized config based on the GUI state.
        """
        if self.source == "device":
            # Gather device name/entry
            device_name = ""
            device_entry = ""
            if hasattr(self, "device_edit"):
                device_name = self.device_edit.text()
            if hasattr(self, "entry_edit"):
                device_entry = self.entry_validator.validate_signal(
                    name=device_name, entry=self.entry_edit.text()
                )
                self.entry_edit.setText(device_entry)
            self.config.signal = DeviceSignal(name=device_name, entry=device_entry)
            self.config.source = "device"
            self.config.label = f"{device_name}-{device_entry}"
        else:
            # DAP logic
            parent_conf_dict = {}
            if self.parent_item:
                parent_conf_dict = self.parent_item.export_data()
            parent_conf = CurveConfig(**parent_conf_dict)
            dev_name = ""
            dev_entry = ""
            if parent_conf.signal:
                dev_name = parent_conf.signal.name
                dev_entry = parent_conf.signal.entry
            # Dap from the DapComboBox
            new_dap = "GaussianModel"
            if hasattr(self, "dap_combo"):
                new_dap = self.dap_combo.fit_model_combobox.currentText()
            self.config.signal = DeviceSignal(name=dev_name, entry=dev_entry, dap=new_dap)
            self.config.source = "dap"
            self.config.parent_label = parent_conf.label
            self.config.label = f"{parent_conf.label}-{new_dap}"

        # Common style fields
        self.config.color = self.color_button.color
        self.config.symbol_color = self.color_button.color
        self.config.pen_style = self.style_combo.currentText()
        self.config.pen_width = self.width_spin.value()
        self.config.symbol_size = self.symbol_spin.value()

        return self.config.model_dump()

    def closeEvent(self, event) -> None:
        logger.info(f"CurveRow closeEvent: {self.config.label}")
        return super().closeEvent(event)


class CurveTree(BECWidget, QWidget):
    """A tree widget that manages device and DAP curves."""

    PLUGIN = False
    RPC = False

    def __init__(
        self,
        parent: QWidget | None = None,
        config: ConnectionConfig | None = None,
        client=None,
        gui_id: str | None = None,
        waveform: Waveform | None = None,
        **kwargs,
    ) -> None:
        if config is None:
            config = ConnectionConfig(widget_class=self.__class__.__name__)
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.waveform = waveform
        if self.waveform and hasattr(self.waveform, "color_palette"):
            self.color_palette = self.waveform.color_palette
        else:
            self.color_palette = "plasma"

        self.get_bec_shortcuts()

        self.color_buffer = []
        self.all_items = []
        self.layout = QVBoxLayout(self)
        self._init_toolbar()
        self._init_tree()
        self.refresh_from_waveform()

    def _init_toolbar(self):
        """Initialize the toolbar with actions: add, send, refresh, expand, collapse, renormalize."""
        self.toolbar = ModularToolBar(parent=self, target_widget=self, orientation="horizontal")
        add = MaterialIconAction(
            icon_name="add", tooltip="Add new curve", checkable=False, parent=self
        )
        expand = MaterialIconAction(
            icon_name="unfold_more", tooltip="Expand All DAP", checkable=False, parent=self
        )
        collapse = MaterialIconAction(
            icon_name="unfold_less", tooltip="Collapse All DAP", checkable=False, parent=self
        )

        self.toolbar.add_action("add", add, self)
        self.toolbar.add_action("expand_all", expand, self)
        self.toolbar.add_action("collapse_all", collapse, self)

        # Add colormap widget (not updating waveform's color_palette until Send is pressed)
        self.spacer = QWidget()
        self.spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.toolbar.addWidget(self.spacer)

        # Renormalize colors button
        renorm_action = MaterialIconAction(
            icon_name="palette", tooltip="Normalize All Colors", checkable=False, parent=self
        )
        self.toolbar.add_action("renormalize_colors", renorm_action, self)
        renorm_action.action.triggered.connect(lambda checked: self.renormalize_colors())

        self.colormap_widget = BECColorMapWidget(cmap=self.color_palette or "plasma")
        self.toolbar.addWidget(self.colormap_widget)
        self.colormap_widget.colormap_changed_signal.connect(self.handle_colormap_changed)

        add.action.triggered.connect(lambda checked: self.add_new_curve())
        expand.action.triggered.connect(lambda checked: self.expand_all_daps())
        collapse.action.triggered.connect(lambda checked: self.collapse_all_daps())

        self.layout.addWidget(self.toolbar)

    def _init_tree(self):
        """Initialize the QTreeWidget with 7 columns and compact widths."""
        self.tree = QTreeWidget()
        self.tree.setColumnCount(7)
        self.tree.setHeaderLabels(["Actions", "Name", "Entry", "Color", "Style", "Width", "Symbol"])
        self.tree.setColumnWidth(0, 90)
        self.tree.setColumnWidth(1, 100)
        self.tree.setColumnWidth(2, 100)
        self.tree.setColumnWidth(3, 70)
        self.tree.setColumnWidth(4, 80)
        self.tree.setColumnWidth(5, 40)
        self.tree.setColumnWidth(6, 40)
        self.layout.addWidget(self.tree)

    def _init_color_buffer(self, size: int):
        """
        Initializes the color buffer with a calculated set of colors based on the golden
        angle sequence.

        Args:
            size (int): The number of colors to be generated for the color buffer.
        """
        self.color_buffer = Colors.golden_angle_color(
            colormap=self.colormap_widget.colormap, num=size, format="HEX"
        )

    def _ensure_color_buffer_size(self):
        """
        Ensures that the color buffer size meets the required number of items.
        """
        current_count = len(self.color_buffer)
        color_list = Colors.golden_angle_color(
            colormap=self.color_palette, num=max(10, current_count + 1), format="HEX"
        )
        self.color_buffer = color_list

    def handle_colormap_changed(self, new_cmap: str):
        """
        Handles the updating of the color palette when the colormap is changed.

        Args:
            new_cmap: The new colormap to be set as the color palette.
        """
        self.color_palette = new_cmap

    def renormalize_colors(self):
        """Overwrite all existing rows with new colors from the buffer in their creation order."""
        total = len(self.all_items)
        self._ensure_color_buffer_size()
        for idx, item in enumerate(self.all_items):
            if hasattr(item, "color_button"):
                new_col = self.color_buffer[idx]
                item.color_button.set_color(new_col)
                if hasattr(item, "config"):
                    item.config.color = new_col
                    item.config.symbol_color = new_col

    def add_new_curve(self, name: str = None, entry: str = None):
        """Add a new device-type CurveRow with an assigned colormap color.

        Args:
            name (str, optional): Device name.
            entry (str, optional): Device entry.
            style (str, optional): Pen style. Defaults to "solid".
            width (int, optional): Pen width. Defaults to 4.
            symbol_size (int, optional): Symbol size. Defaults to 7.

        Returns:
            CurveRow: The newly created top-level row.
        """
        cfg = CurveConfig(
            widget_class="Curve",
            parent_id=self.waveform.gui_id,
            source="device",
            signal=DeviceSignal(name=name or "", entry=entry or ""),
        )
        new_row = CurveRow(self.tree, parent_item=None, config=cfg, device_manager=self.dev)

        # Assign color from the buffer ONLY to this new curve.
        total_items = len(self.all_items)
        self._ensure_color_buffer_size()
        color_idx = total_items - 1  # new row is last
        new_col = self.color_buffer[color_idx]
        new_row.color_button.set_color(new_col)
        new_row.config.color = new_col
        new_row.config.symbol_color = new_col

        return new_row

    def send_curve_json(self):
        """Send the current tree's config as JSON to the waveform, updating wavefrom.color_palette as well."""
        if self.waveform is not None:
            self.waveform.color_palette = self.color_palette
        data = self.export_all_curves()
        json_data = json.dumps(data, indent=2)
        if self.waveform is not None:
            self.waveform.curve_json = json_data

    def export_all_curves(self) -> list:
        """Recursively export data from each row.

        Returns:
            list: A list of exported config dicts for every row (device and DAP).
        """
        curves = []
        for i in range(self.tree.topLevelItemCount()):
            item = self.tree.topLevelItem(i)
            if isinstance(item, CurveRow):
                curves.append(item.export_data())
                for j in range(item.childCount()):
                    child = item.child(j)
                    if isinstance(child, CurveRow):
                        curves.append(child.export_data())
        return curves

    def expand_all_daps(self):
        """Expand all top-level rows to reveal child DAP rows."""
        for i in range(self.tree.topLevelItemCount()):
            item = self.tree.topLevelItem(i)
            self.tree.expandItem(item)

    def collapse_all_daps(self):
        """Collapse all top-level rows, hiding child DAP rows."""
        for i in range(self.tree.topLevelItemCount()):
            item = self.tree.topLevelItem(i)
            self.tree.collapseItem(item)

    def refresh_from_waveform(self):
        """Clear the tree and rebuild from the waveform's existing curves if any, else add sample rows."""
        if self.waveform is None:
            return
        self.tree.clear()
        self.all_items = []

        device_curves = [c for c in self.waveform.curves if c.config.source == "device"]
        dap_curves = [c for c in self.waveform.curves if c.config.source == "dap"]
        for dev in device_curves:
            dr = CurveRow(self.tree, parent_item=None, config=dev.config, device_manager=self.dev)
            for dap in dap_curves:
                if dap.config.parent_label == dev.config.label:
                    CurveRow(self.tree, parent_item=dr, config=dap.config, device_manager=self.dev)

    def cleanup(self):
        """Cleanup the widget."""
        all_items = list(self.all_items)
        for item in all_items:
            item.remove_self()

    def closeEvent(self, event):
        self.cleanup()
        return super().closeEvent(event)



from __future__ import annotations

from bec_lib import bec_logger
from qtpy.QtCore import QSettings
from qtpy.QtWidgets import (
    QApplication,
    QCheckBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

logger = bec_logger.logger


class WidgetStateManager:
    """
    A class to manage the state of a widget by saving and loading the state to and from a INI file.

    Args:
        widget(QWidget): The widget to manage the state for.
    """

    def __init__(self, widget):
        self.widget = widget

    def save_state(self, filename: str = None):
        """
        Save the state of the widget to an INI file.

        Args:
            filename(str): The filename to save the state to.
        """
        if not filename:
            filename, _ = QFileDialog.getSaveFileName(
                self.widget, "Save Settings", "", "INI Files (*.ini)"
            )
        if filename:
            settings = QSettings(filename, QSettings.IniFormat)
            self._save_widget_state_qsettings(self.widget, settings)

    def load_state(self, filename: str = None):
        """
        Load the state of the widget from an INI file.

        Args:
            filename(str): The filename to load the state from.
        """
        if not filename:
            filename, _ = QFileDialog.getOpenFileName(
                self.widget, "Load Settings", "", "INI Files (*.ini)"
            )
        if filename:
            settings = QSettings(filename, QSettings.IniFormat)
            self._load_widget_state_qsettings(self.widget, settings)

    def _save_widget_state_qsettings(self, widget: QWidget, settings: QSettings):
        """
        Save the state of the widget to QSettings.

        Args:
            widget(QWidget): The widget to save the state for.
            settings(QSettings): The QSettings object to save the state to.
        """
        if widget.property("skip_settings") is True:
            return

        meta = widget.metaObject()
        widget_name = self._get_full_widget_name(widget)
        settings.beginGroup(widget_name)
        for i in range(meta.propertyCount()):
            prop = meta.property(i)
            name = prop.name()
            if (
                name == "objectName"
                or not prop.isReadable()
                or not prop.isWritable()
                or not prop.isStored()  # can be extended to fine filter
            ):
                continue
            value = widget.property(name)
            settings.setValue(name, value)
        settings.endGroup()

        # Recursively process children (only if they aren't skipped)
        for child in widget.children():
            if (
                child.objectName()
                and child.property("skip_settings") is not True
                and not isinstance(child, QLabel)
            ):
                self._save_widget_state_qsettings(child, settings)

    def _load_widget_state_qsettings(self, widget: QWidget, settings: QSettings):
        """
        Load the state of the widget from QSettings.

        Args:
            widget(QWidget): The widget to load the state for.
            settings(QSettings): The QSettings object to load the state from.
        """
        if widget.property("skip_settings") is True:
            return

        meta = widget.metaObject()
        widget_name = self._get_full_widget_name(widget)
        settings.beginGroup(widget_name)
        for i in range(meta.propertyCount()):
            prop = meta.property(i)
            name = prop.name()
            if settings.contains(name):
                value = settings.value(name)
                widget.setProperty(name, value)
        settings.endGroup()

        # Recursively process children (only if they aren't skipped)
        for child in widget.children():
            if (
                child.objectName()
                and child.property("skip_settings") is not True
                and not isinstance(child, QLabel)
            ):
                self._load_widget_state_qsettings(child, settings)

    def _get_full_widget_name(self, widget: QWidget):
        """
        Get the full name of the widget including its parent names.

        Args:
            widget(QWidget): The widget to get the full name for.

        Returns:
            str: The full name of the widget.
        """
        name = widget.objectName()
        parent = widget.parent()
        while parent:
            obj_name = parent.objectName() or parent.metaObject().className()
            name = obj_name + "." + name
            parent = parent.parent()
        return name


class ExampleApp(QWidget):  # pragma: no cover:
    def __init__(self):
        super().__init__()
        self.setObjectName("MainWindow")
        self.setWindowTitle("State Manager Example")

        layout = QVBoxLayout(self)

        # A line edit to store some user text
        self.line_edit = QLineEdit(self)
        self.line_edit.setObjectName("MyLineEdit")
        self.line_edit.setPlaceholderText("Enter some text here...")
        layout.addWidget(self.line_edit)

        # A spin box to hold a numeric value
        self.spin_box = QSpinBox(self)
        self.spin_box.setObjectName("MySpinBox")
        self.spin_box.setRange(0, 100)
        layout.addWidget(self.spin_box)

        # A checkbox to hold a boolean value
        self.check_box = QCheckBox("Enable feature?", self)
        self.check_box.setObjectName("MyCheckBox")
        layout.addWidget(self.check_box)

        # A checkbox that we want to skip
        self.check_box_skip = QCheckBox("Enable feature - skip save?", self)
        self.check_box_skip.setProperty("skip_state", True)
        self.check_box_skip.setObjectName("MyCheckBoxSkip")
        layout.addWidget(self.check_box_skip)

        #  CREATE A "SIDE PANEL" with nested structure and skip all what is inside
        self.side_panel = QWidget(self)
        self.side_panel.setObjectName("SidePanel")
        self.side_panel.setProperty("skip_settings", True)  # skip the ENTIRE panel
        layout.addWidget(self.side_panel)

        # Put some sub-widgets inside side_panel
        panel_layout = QVBoxLayout(self.side_panel)
        self.panel_label = QLabel("Label in side panel", self.side_panel)
        self.panel_label.setObjectName("PanelLabel")
        panel_layout.addWidget(self.panel_label)

        self.panel_edit = QLineEdit(self.side_panel)
        self.panel_edit.setObjectName("PanelLineEdit")
        self.panel_edit.setPlaceholderText("I am inside side panel")
        panel_layout.addWidget(self.panel_edit)

        self.panel_checkbox = QCheckBox("Enable feature in side panel?", self.side_panel)
        self.panel_checkbox.setObjectName("PanelCheckBox")
        panel_layout.addWidget(self.panel_checkbox)

        # Save/Load buttons
        button_layout = QHBoxLayout()
        self.save_button = QPushButton("Save State", self)
        self.load_button = QPushButton("Load State", self)
        button_layout.addWidget(self.save_button)
        button_layout.addWidget(self.load_button)
        layout.addLayout(button_layout)

        # Create the state manager
        self.state_manager = WidgetStateManager(self)

        # Connect buttons
        self.save_button.clicked.connect(lambda: self.state_manager.save_state())
        self.load_button.clicked.connect(lambda: self.state_manager.load_state())


if __name__ == "__main__":  # pragma: no cover:
    import sys

    app = QApplication(sys.argv)
    w = ExampleApp()
    w.show()
    sys.exit(app.exec_())



import traceback

from pyqtgraph.exporters import MatplotlibExporter

from bec_widgets.utils.error_popups import SafeSlot, WarningPopupUtility
from bec_widgets.utils.toolbar import MaterialIconAction, SwitchableToolBarAction, ToolbarBundle


class PlotExportBundle(ToolbarBundle):
    """
    A bundle of actions that are hooked in this constructor itself,
    so that you can immediately connect the signals and toggle states.

    This bundle is for a toolbar that controls exporting a plot.
    """

    def __init__(self, bundle_id="mouse_interaction", target_widget=None, **kwargs):
        super().__init__(bundle_id=bundle_id, actions=[], **kwargs)
        self.target_widget = target_widget

        # Create each MaterialIconAction with a parent
        # so the signals can fire even if the toolbar isn't added yet.
        save = MaterialIconAction(
            icon_name="save", tooltip="Open Export Dialog", parent=self.target_widget
        )
        matplotlib = MaterialIconAction(
            icon_name="photo_library", tooltip="Open Matplotlib Dialog", parent=self.target_widget
        )

        switch_export_action = SwitchableToolBarAction(
            actions={"save": save, "matplotlib": matplotlib},
            initial_action="save",
            tooltip="Switchable Action",
            checkable=False,
            parent=self,
        )

        # Add them to the bundle
        self.add_action("export_switch", switch_export_action)

        # Immediately connect signals
        save.action.triggered.connect(self.export_dialog)
        matplotlib.action.triggered.connect(self.matplotlib_dialog)

    @SafeSlot()
    def export_dialog(self):
        """
        Open the export dialog for the plot widget.
        """
        if self.target_widget:
            scene = self.target_widget.plot_item.scene()
            scene.contextMenuItem = self.target_widget.plot_item
            scene.showExportDialog()

    @SafeSlot()
    def matplotlib_dialog(self):
        """
        Export the plot widget to Matplotlib.
        """
        if self.target_widget:
            try:
                import matplotlib as mpl

                MatplotlibExporter(self.target_widget.plot_item).export()
            except ModuleNotFoundError:
                warning_util = WarningPopupUtility()
                warning_util.show_warning(
                    title="Matplotlib not installed",
                    message="Matplotlib is required for this feature.",
                    detailed_text="Please install matplotlib in your Python environment by using 'pip install matplotlib'.",
                )
                return
            except TypeError:
                warning_util = WarningPopupUtility()
                error_msg = traceback.format_exc()
                warning_util.show_warning(
                    title="Matplotlib TypeError",
                    message="Matplotlib exporter could not resolve the plot item.",
                    detailed_text=error_msg,
                )
                return



from collections import OrderedDict
from typing import Literal

from qtpy.QtWidgets import QGridLayout, QWidget


class GridLayoutManager:
    """
    GridLayoutManager class is used to manage widgets in a QGridLayout and extend its functionality.

    The GridLayoutManager class provides methods to add, move, and check the position of widgets in a QGridLayout.
    It also provides a method to get the positions of all widgets in the layout.

    Args:
        layout(QGridLayout): The layout to manage.
    """

    def __init__(self, layout: QGridLayout):
        self.layout = layout

    def is_position_occupied(self, row: int, col: int) -> bool:
        """
        Check if the position in the layout is occupied by a widget.

        Args:
            row(int): The row to check.
            col(int): The column to check.

        Returns:
            bool: True if the position is occupied, False otherwise.
        """
        for i in range(self.layout.count()):
            widget_row, widget_col, _, _ = self.layout.getItemPosition(i)
            if widget_row == row and widget_col == col:
                return True
        return False

    def shift_widgets(
        self,
        direction: Literal["down", "up", "left", "right"] = "down",
        start_row: int = 0,
        start_col: int = 0,
    ):
        """
        Shift widgets in the layout in the specified direction starting from the specified position.

        Args:
            direction(str): The direction to shift the widgets. Can be "down", "up", "left", or "right".
            start_row(int): The row to start shifting from. Default is 0.
            start_col(int): The column to start shifting from. Default is 0.
        """
        for i in reversed(range(self.layout.count())):
            widget_item = self.layout.itemAt(i)
            widget = widget_item.widget()
            row, col, rowspan, colspan = self.layout.getItemPosition(i)
            if direction == "down" and row >= start_row:
                self.layout.addWidget(widget, row + 1, col, rowspan, colspan)
            elif direction == "up" and row > start_row:
                self.layout.addWidget(widget, row - 1, col, rowspan, colspan)
            elif direction == "right" and col >= start_col:
                self.layout.addWidget(widget, row, col + 1, rowspan, colspan)
            elif direction == "left" and col > start_col:
                self.layout.addWidget(widget, row, col - 1, rowspan, colspan)

    def move_widget(self, widget: QWidget, new_row: int, new_col: int):
        """
        Move a widget to a new position in the layout.

        Args:
            widget(QWidget): The widget to move.
            new_row(int): The new row to move the widget to.
            new_col(int): The new column to move the widget to.
        """
        self.layout.removeWidget(widget)
        self.layout.addWidget(widget, new_row, new_col)

    def add_widget(
        self,
        widget: QWidget,
        row=None,
        col=0,
        rowspan=1,
        colspan=1,
        shift: Literal["down", "up", "left", "right"] = "down",
    ):
        """
        Add a widget to the layout at the specified position.

        Args:
            widget(QWidget): The widget to add.
            row(int): The row to add the widget to. If None, the widget will be added to the next available row.
            col(int): The column to add the widget to. Default is 0.
            rowspan(int): The number of rows the widget will span. Default is 1.
            colspan(int): The number of columns the widget will span. Default is 1.
            shift(str): The direction to shift the widgets if the position is occupied. Can be "down", "up", "left", or "right".
        """
        if row is None:
            row = self.layout.rowCount()
        if self.is_position_occupied(row, col):
            self.shift_widgets(shift, start_row=row)
        self.layout.addWidget(widget, row, col, rowspan, colspan)

    def get_widgets_positions(self) -> dict:
        """
        Get the positions of all widgets in the layout.
        Returns:
            dict: A dictionary with the positions of the widgets in the layout.

        """
        positions = []
        for i in range(self.layout.count()):
            widget_item = self.layout.itemAt(i)
            widget = widget_item.widget()
            if widget:
                position = self.layout.getItemPosition(i)
                positions.append((position, widget))
        positions.sort(key=lambda x: (x[0][0], x[0][1], x[0][2], x[0][3]))
        ordered_positions = OrderedDict()
        for pos, widget in positions:
            ordered_positions[pos] = widget
        return ordered_positions



from typing import Literal

from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from qtpy.QtCore import Property, Qt, Signal, Slot
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QDoubleSpinBox,
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
)

from bec_widgets.utils.widget_io import WidgetIO
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import BECDeviceFilter
from bec_widgets.widgets.control.device_input.device_line_edit.device_line_edit import (
    DeviceLineEdit,
)

logger = bec_logger.logger


class ScanArgType:
    DEVICE = "device"
    FLOAT = "float"
    INT = "int"
    BOOL = "bool"
    STR = "str"
    DEVICEBASE = "DeviceBase"
    LITERALS = "dict"


class SettingsDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Settings")

        layout = QFormLayout()

        self.precision_spin_box = QSpinBox()
        self.precision_spin_box.setRange(
            -2147483647, 2147483647
        )  # 2147483647 is the largest int which qt allows

        self.step_size_spin_box = QDoubleSpinBox()
        self.step_size_spin_box.setRange(-float("inf"), float("inf"))

        fixed_width = 80
        self.precision_spin_box.setFixedWidth(fixed_width)
        self.step_size_spin_box.setFixedWidth(fixed_width)

        layout.addRow("Decimal Precision:", self.precision_spin_box)
        layout.addRow("Step Size:", self.step_size_spin_box)

        button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)
        layout.addWidget(button_box)

        self.setLayout(layout)

    def getValues(self):
        return self.precision_spin_box.value(), self.step_size_spin_box.value()


class ScanSpinBox(QSpinBox):
    def __init__(
        self, parent=None, arg_name: str = None, default: int | None = None, *args, **kwargs
    ):
        super().__init__(parent=parent, *args, **kwargs)
        self.arg_name = arg_name
        self.setRange(-2147483647, 2147483647)  # 2147483647 is the largest int which qt allows
        if default is not None:
            self.setValue(default)


class ScanDoubleSpinBox(QDoubleSpinBox):
    def __init__(
        self, parent=None, arg_name: str = None, default: float | None = None, *args, **kwargs
    ):
        super().__init__(parent=parent, *args, **kwargs)
        self.arg_name = arg_name
        self.setRange(-float("inf"), float("inf"))
        if default is not None:
            self.setValue(default)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.showSettingsDialog)

        self.setToolTip("Right click to open settings dialog for decimal precision and step size.")

    def showSettingsDialog(self):
        dialog = SettingsDialog(self)
        dialog.precision_spin_box.setValue(self.decimals())
        dialog.step_size_spin_box.setValue(self.singleStep())

        if dialog.exec_() == QDialog.Accepted:
            precision, step_size = dialog.getValues()
            self.setDecimals(precision)
            self.setSingleStep(step_size)


class ScanLineEdit(QLineEdit):
    def __init__(
        self, parent=None, arg_name: str = None, default: str | None = None, *args, **kwargs
    ):
        super().__init__(parent=parent, *args, **kwargs)
        self.arg_name = arg_name
        if default is not None:
            self.setText(default)


class ScanCheckBox(QCheckBox):
    def __init__(
        self, parent=None, arg_name: str = None, default: bool | None = None, *args, **kwargs
    ):
        super().__init__(parent=parent, *args, **kwargs)
        self.arg_name = arg_name
        if default is not None:
            self.setChecked(default)


class ScanGroupBox(QGroupBox):
    WIDGET_HANDLER = {
        ScanArgType.DEVICE: DeviceLineEdit,
        ScanArgType.DEVICEBASE: DeviceLineEdit,
        ScanArgType.FLOAT: ScanDoubleSpinBox,
        ScanArgType.INT: ScanSpinBox,
        ScanArgType.BOOL: ScanCheckBox,
        ScanArgType.STR: ScanLineEdit,
        ScanArgType.LITERALS: QComboBox,  # TODO figure out combobox logic
    }

    device_selected = Signal(str)

    def __init__(
        self,
        parent=None,
        box_type=Literal["args", "kwargs"],
        config: dict | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(parent=parent, *args, **kwargs)
        self.config = config
        self.box_type = box_type
        self._hide_add_remove_buttons = False

        vbox_layout = QVBoxLayout(self)
        hbox_layout = QHBoxLayout()
        vbox_layout.addLayout(hbox_layout)
        self.layout = QGridLayout(self)
        vbox_layout.addLayout(self.layout)

        # Add bundle button
        self.button_add_bundle = QPushButton(self)
        self.button_add_bundle.setIcon(
            material_icon(icon_name="add", size=(15, 15), convert_to_pixmap=False)
        )
        # Remove bundle button
        self.button_remove_bundle = QPushButton(self)
        self.button_remove_bundle.setIcon(
            material_icon(icon_name="remove", size=(15, 15), convert_to_pixmap=False)
        )
        hbox_layout.addWidget(self.button_add_bundle)
        hbox_layout.addWidget(self.button_remove_bundle)

        self.labels = []
        self.widgets = []
        self.selected_devices = {}

        self.init_box(self.config)

        self.button_add_bundle.clicked.connect(self.add_widget_bundle)
        self.button_remove_bundle.clicked.connect(self.remove_widget_bundle)

    def init_box(self, config: dict):
        box_name = config.get("name", "ScanGroupBox")
        self.inputs = config.get("inputs", {})
        self.setTitle(box_name)

        # Labels
        self.add_input_labels(self.inputs, 0)

        # Widgets
        if self.box_type == "args":
            min_bundle = self.config.get("min", 1)
            for i in range(1, min_bundle + 1):
                self.add_input_widgets(self.inputs, i)
        else:
            self.add_input_widgets(self.inputs, 1)
            self.button_add_bundle.setVisible(False)
            self.button_remove_bundle.setVisible(False)

    def add_input_labels(self, group_inputs: dict, row: int) -> None:
        """
        Adds the given arg_group from arg_bundle to the scan control layout. The input labels are always added to the first row.

        Args:
            group(dict): Dictionary containing the arg_group information.
        """
        for column_index, item in enumerate(group_inputs):
            arg_name = item.get("name", None)
            display_name = item.get("display_name", arg_name)
            label = QLabel(text=display_name)
            self.layout.addWidget(label, row, column_index)
            self.labels.append(label)

    def add_input_widgets(self, group_inputs: dict, row) -> None:
        """
        Adds the given arg_group from arg_bundle to the scan control layout.

        Args:
            group_inputs(dict): Dictionary containing the arg_group information.
            row(int): The row to add the widgets to.
        """
        for column_index, item in enumerate(group_inputs):
            arg_name = item.get("name", None)
            default = item.get("default", None)
            widget_class = self.WIDGET_HANDLER.get(item["type"], None)
            if widget_class is None:
                logger.error(
                    f"Unsupported annotation '{item['type']}' for parameter '{item['name']}'"
                )
                continue
            if default == "_empty":
                default = None
            widget = widget_class(parent=self.parent(), arg_name=arg_name, default=default)
            if isinstance(widget, DeviceLineEdit):
                widget.set_device_filter(BECDeviceFilter.DEVICE)
                self.selected_devices[widget] = ""
                widget.device_selected.connect(self.emit_device_selected)
            tooltip = item.get("tooltip", None)
            if tooltip is not None:
                widget.setToolTip(item["tooltip"])
            self.layout.addWidget(widget, row, column_index)
            self.widgets.append(widget)

    @Slot(str)
    def emit_device_selected(self, device_name):
        self.selected_devices[self.sender()] = device_name.strip()
        selected_devices_str = " ".join(self.selected_devices.values())
        self.device_selected.emit(selected_devices_str)

    def add_widget_bundle(self):
        """
        Adds a new row of widgets to the scan control layout. Only usable for arg_groups.
        """
        arg_max = self.config.get("max", None)
        row = self.layout.rowCount()
        if arg_max is not None and row >= arg_max:
            return

        self.add_input_widgets(self.inputs, row)

    def remove_widget_bundle(self):
        """
        Removes the last row of widgets from the scan control layout. Only usable for arg_groups.
        """
        arg_min = self.config.get("min", None)
        row = self.count_arg_rows()
        if arg_min is not None and row <= arg_min:
            return

        for widget in self.widgets[-len(self.inputs) :]:
            if isinstance(widget, DeviceLineEdit):
                self.selected_devices[widget] = ""
            widget.close()
            widget.deleteLater()
        self.widgets = self.widgets[: -len(self.inputs)]

        selected_devices_str = " ".join(self.selected_devices.values())
        self.device_selected.emit(selected_devices_str.strip())

    def remove_all_widget_bundles(self):
        """Remove every widget bundle from the scan control layout."""
        for widget in list(self.widgets):
            if isinstance(widget, DeviceLineEdit):
                self.selected_devices.pop(widget, None)
            widget.close()
            widget.deleteLater()
            self.layout.removeWidget(widget)
        self.widgets.clear()
        self.device_selected.emit("")

    @Property(bool)
    def hide_add_remove_buttons(self):
        return self._hide_add_remove_buttons

    @hide_add_remove_buttons.setter
    def hide_add_remove_buttons(self, hide: bool):
        self._hide_add_remove_buttons = hide
        if not hide and self.box_type == "args":
            self.button_add_bundle.show()
            self.button_remove_bundle.show()
            return
        self.button_add_bundle.hide()
        self.button_remove_bundle.hide()

    def get_parameters(self, device_object: bool = True):
        """
        Returns the parameters from the widgets in the scan control layout formated to run scan from BEC.
        """
        if self.box_type == "args":
            return self._get_arg_parameterts(device_object=device_object)
        elif self.box_type == "kwargs":
            return self._get_kwarg_parameters(device_object=device_object)

    def _get_arg_parameterts(self, device_object: bool = True):
        args = []
        for i in range(1, self.layout.rowCount()):
            for j in range(self.layout.columnCount()):
                try:  # In case that the bundle size changes
                    widget = self.layout.itemAtPosition(i, j).widget()
                    if isinstance(widget, DeviceLineEdit) and device_object:
                        value = widget.get_current_device()
                    else:
                        value = WidgetIO.get_value(widget)
                    args.append(value)
                except AttributeError:
                    continue
        return args

    def _get_kwarg_parameters(self, device_object: bool = True):
        kwargs = {}
        for i in range(self.layout.columnCount()):
            widget = self.layout.itemAtPosition(1, i).widget()
            if isinstance(widget, DeviceLineEdit) and device_object:
                value = widget.get_current_device().name
            else:
                value = WidgetIO.get_value(widget)
            kwargs[widget.arg_name] = value
        return kwargs

    def count_arg_rows(self):
        widget_rows = 0
        for row in range(self.layout.rowCount()):
            for col in range(self.layout.columnCount()):
                item = self.layout.itemAtPosition(row, col)
                if item is not None:
                    widget = item.widget()
                    if widget is not None:
                        if isinstance(widget, DeviceLineEdit):
                            widget_rows += 1
        return widget_rows

    def set_parameters(self, parameters: list | dict):
        if self.box_type == "args":
            self._set_arg_parameters(parameters)
        elif self.box_type == "kwargs":
            self._set_kwarg_parameters(parameters)

    def _set_arg_parameters(self, parameters: list):
        self.remove_all_widget_bundles()
        if not parameters:
            return

        inputs_per_bundle = len(self.inputs)
        if inputs_per_bundle == 0:
            return

        bundles_needed = -(-len(parameters) // inputs_per_bundle)

        for row in range(1, bundles_needed + 1):
            self.add_input_widgets(self.inputs, row)

        for i, value in enumerate(parameters):
            WidgetIO.set_value(self.widgets[i], value)

    def _set_kwarg_parameters(self, parameters: dict):
        for widget in self.widgets:
            for key, value in parameters.items():
                if widget.arg_name == key:
                    WidgetIO.set_value(widget, value)
                    break



from bec_lib.callback_handler import EventType
from bec_lib.device import Signal
from bec_lib.logger import bec_logger
from qtpy.QtCore import Property, Slot

from bec_widgets.utils import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.filter_io import FilterIO
from bec_widgets.utils.ophyd_kind_util import Kind
from bec_widgets.utils.widget_io import WidgetIO

logger = bec_logger.logger


class DeviceSignalInputBaseConfig(ConnectionConfig):
    """Configuration class for DeviceSignalInputBase."""

    signal_filter: str | list[str] | None = None
    default: str | None = None
    arg_name: str | None = None
    device: str | None = None
    signals: list[str] | None = None


class DeviceSignalInputBase(BECWidget):
    """
    Mixin base class for device signal input widgets.
    Mixin class for device signal input widgets. This class provides methods to get the device signal list and device
    signal object based on the current text of the widget.
    """

    RPC = False
    _filter_handler = {
        Kind.hinted: "include_hinted_signals",
        Kind.normal: "include_normal_signals",
        Kind.config: "include_config_signals",
    }

    def __init__(
        self,
        client=None,
        config: DeviceSignalInputBaseConfig | dict | None = None,
        gui_id: str = None,
        **kwargs,
    ):

        self.config = self._process_config_input(config)
        super().__init__(client=client, config=self.config, gui_id=gui_id, **kwargs)

        self._device = None
        self.get_bec_shortcuts()
        self._signal_filter = []
        self._signals = []
        self._hinted_signals = []
        self._normal_signals = []
        self._config_signals = []
        self.bec_dispatcher.client.callbacks.register(
            EventType.DEVICE_UPDATE, self.update_signals_from_filters
        )

    ### Qt Slots ###

    @Slot(str)
    def set_signal(self, signal: str):
        """
        Set the signal.

        Args:
            signal (str): signal name.
        """
        if self.validate_signal(signal) is True:
            WidgetIO.set_value(widget=self, value=signal)
            self.config.default = signal
        else:
            logger.warning(
                f"Signal {signal} not found for device {self.device} and filtered selection {self.signal_filter}."
            )

    @Slot(str)
    def set_device(self, device: str | None):
        """
        Set the device. If device is not valid, device will be set to None which happpens

        Args:
            device(str): device name.
        """
        if self.validate_device(device) is False:
            self._device = None
        else:
            self._device = device
        self.update_signals_from_filters()

    @Slot(dict, dict)
    @Slot()
    def update_signals_from_filters(
        self, content: dict | None = None, metadata: dict | None = None
    ):
        """Update the filters for the device signals based on list in self.signal_filter.
        In addition, store the hinted, normal and config signals in separate lists to allow
        customisation within QLineEdit.

        Note:
        Signal and ComputedSignals have no signals. The naming convention follows the device name.
        """
        self.config.signal_filter = self.signal_filter
        # pylint: disable=protected-access
        if not self.validate_device(self._device):
            self._device = None
            self.config.device = self._device
            return
        device = self.get_device_object(self._device)
        # See above convention for Signals and ComputedSignals
        if isinstance(device, Signal):
            self._signals = [self._device]
            FilterIO.set_selection(widget=self, selection=[self._device])
            return
        device_info = device._info["signals"]

        def _update(kind: Kind):
            return [
                signal
                for signal, signal_info in device_info.items()
                if kind in self.signal_filter
                and (signal_info.get("kind_str", None) == str(kind.name))
            ]

        self._hinted_signals = _update(Kind.hinted)
        self._normal_signals = _update(Kind.normal)
        self._config_signals = _update(Kind.config)

        self._signals = self._hinted_signals + self._normal_signals + self._config_signals
        FilterIO.set_selection(widget=self, selection=self.signals)

    ### Qt Properties ###

    @Property(str)
    def device(self) -> str:
        """Get the selected device."""
        if self._device is None:
            return ""
        return self._device

    @device.setter
    def device(self, value: str):
        """Set the device and update the filters, only allow devices present in the devicemanager."""
        self._device = value
        self.config.device = value
        self.update_signals_from_filters()

    @Property(bool)
    def include_hinted_signals(self):
        """Include hinted signals in filters."""
        return Kind.hinted in self.signal_filter

    @include_hinted_signals.setter
    def include_hinted_signals(self, value: bool):
        if value:
            self._signal_filter.append(Kind.hinted)
        else:
            self._signal_filter.remove(Kind.hinted)
        self.update_signals_from_filters()

    @Property(bool)
    def include_normal_signals(self):
        """Include normal signals in filters."""
        return Kind.normal in self.signal_filter

    @include_normal_signals.setter
    def include_normal_signals(self, value: bool):
        if value:
            self._signal_filter.append(Kind.normal)
        else:
            self._signal_filter.remove(Kind.normal)
        self.update_signals_from_filters()

    @Property(bool)
    def include_config_signals(self):
        """Include config signals in filters."""
        return Kind.config in self.signal_filter

    @include_config_signals.setter
    def include_config_signals(self, value: bool):
        if value:
            self._signal_filter.append(Kind.config)
        else:
            self._signal_filter.remove(Kind.config)
        self.update_signals_from_filters()

    ### Properties and Methods ###

    @property
    def signals(self) -> list[str]:
        """
        Get the list of device signals for the applied filters.

        Returns:
            list[str]: List of device signals.
        """
        return self._signals

    @signals.setter
    def signals(self, value: list[str]):
        self._signals = value
        self.config.signals = value
        FilterIO.set_selection(widget=self, selection=value)

    @property
    def signal_filter(self) -> list[str]:
        """Get the list of filters to apply on the device signals."""
        return self._signal_filter

    def get_available_filters(self) -> list[str]:
        """Get the available filters."""
        return [entry for entry in self._filter_handler]

    def set_filter(self, filter_selection: str | list[str]):
        """
        Set the device filter. If None, all devices are included.

        Args:
            filter_selection (str | list[str]): Device filters from BECDeviceFilter and BECReadoutPriority.
        """
        filters = None
        if isinstance(filter_selection, list):
            filters = [self._filter_handler.get(entry) for entry in filter_selection]
        if isinstance(filter_selection, str):
            filters = [self._filter_handler.get(filter_selection)]
        if filters is None:
            return
        for entry in filters:
            setattr(self, entry, True)

    def get_device_object(self, device: str) -> object | None:
        """
        Get the device object based on the device name.

        Args:
            device(str): Device name.

        Returns:
            object: Device object, can be device of type Device, Positioner, Signal or ComputedSignal.
        """
        self.validate_device(device)
        dev = getattr(self.dev, device, None)
        if dev is None:
            logger.warning(f"Device {device} not found in devicemanager.")
            return None
        return dev

    def validate_device(self, device: str | None, raise_on_false: bool = False) -> bool:
        """
        Validate the device if it is present in current BEC instance.

        Args:
            device(str): Device to validate.
        """
        if device in self.dev:
            return True
        if raise_on_false is True:
            raise ValueError(f"Device {device} not found in devicemanager.")
        return False

    def validate_signal(self, signal: str) -> bool:
        """
        Validate the signal if it is present in the device signals.

        Args:
            signal(str): Signal to validate.
        """
        if signal in self.signals:
            return True
        return False

    def _process_config_input(self, config: DeviceSignalInputBaseConfig | dict | None):
        if config is None:
            return DeviceSignalInputBaseConfig(widget_class=self.__class__.__name__)
        return DeviceSignalInputBaseConfig.model_validate(config)



from bec_lib.logger import bec_logger
from PySide6.QtGui import QCloseEvent
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QPushButton, QVBoxLayout, QWidget

from bec_widgets.utils.error_popups import SafeSlot

logger = bec_logger.logger


class SettingWidget(QWidget):
    """
    Abstract class for a settings widget to enforce the implementation of the accept_changes and display_current_settings.
    Can be used for toolbar actions to display the settings of a widget.

    Args:
        target_widget (QWidget): The widget that the settings will be taken from and applied to.
    """

    def __init__(self, parent=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        self.target_widget = None

    def set_target_widget(self, target_widget: QWidget):
        self.target_widget = target_widget

    @SafeSlot()
    def accept_changes(self):
        """
        Accepts the changes made in the settings widget and applies them to the target widget.
        """
        pass

    @SafeSlot(dict)
    def display_current_settings(self, config_dict: dict):
        """
        Displays the current settings of the target widget in the settings widget.

        Args:
            config_dict(dict): The current settings of the target widget.
        """
        pass

    def cleanup(self):
        """
        Cleanup the settings widget.
        """

    def closeEvent(self, event: QCloseEvent) -> None:
        self.cleanup()
        return super().closeEvent(event)


class SettingsDialog(QDialog):
    """
    Dialog to display and edit the settings of a widget with accept and cancel buttons.

    Args:
        parent (QWidget): The parent widget of the dialog.
        target_widget (QWidget): The widget that the settings will be taken from and applied to.
        settings_widget (SettingWidget): The widget that will display the settings.
    """

    def __init__(
        self,
        parent=None,
        settings_widget: SettingWidget = None,
        window_title: str = "Settings",
        config: dict = None,
        modal: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(parent, *args, **kwargs)

        self.setModal(modal)

        self.setWindowTitle(window_title)

        self.widget = settings_widget
        self.widget.set_target_widget(parent)
        if config is None:
            config = parent.get_config()

        self.widget.display_current_settings(config)

        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        self.apply_button = QPushButton("Apply")

        button_layout = QHBoxLayout()
        button_layout.addWidget(self.button_box.button(QDialogButtonBox.Cancel))
        button_layout.addWidget(self.apply_button)
        button_layout.addWidget(self.button_box.button(QDialogButtonBox.Ok))

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        self.apply_button.clicked.connect(self.apply_changes)

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(5, 5, 5, 5)
        self.layout.addWidget(self.widget)
        self.layout.addLayout(button_layout)

        ok_button = self.button_box.button(QDialogButtonBox.Ok)
        ok_button.setDefault(True)
        ok_button.setAutoDefault(True)

    @SafeSlot()
    def accept(self):
        """
        Accept the changes made in the settings widget and close the dialog.
        """
        self.widget.accept_changes()
        self.cleanup()
        super().accept()

    @SafeSlot()
    def reject(self):
        """
        Reject the changes made in the settings widget and close the dialog.
        """
        self.cleanup()
        super().reject()

    @SafeSlot()
    def apply_changes(self):
        """
        Apply the changes made in the settings widget without closing the dialog.
        """
        self.widget.accept_changes()

    def cleanup(self):
        """
        Cleanup the dialog.
        """
        self.button_box.close()
        self.button_box.deleteLater()
        self.widget.close()
        self.widget.deleteLater()

    def closeEvent(self, event):
        logger.info("Closing settings dialog")
        self.cleanup()
        super().closeEvent(event)



import sys
from string import Template

from qtpy.QtCore import Property, QEasingCurve, QPropertyAnimation, QRectF, Qt, QTimer, Slot
from qtpy.QtGui import QColor, QPainter, QPainterPath
from qtpy.QtWidgets import QApplication, QLabel, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors


class BECProgressBar(BECWidget, QWidget):
    """
    A custom progress bar with smooth transitions. The displayed text can be customized using a template.
    """

    PLUGIN = True
    USER_ACCESS = [
        "set_value",
        "set_maximum",
        "set_minimum",
        "label_template",
        "label_template.setter",
        "_get_label",
    ]
    ICON_NAME = "page_control"

    def __init__(self, parent=None, client=None, config=None, gui_id=None, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        accent_colors = get_accent_colors()

        # internal values
        self._oversampling_factor = 50
        self._value = 0
        self._target_value = 0
        self._maximum = 100 * self._oversampling_factor

        # User values
        self._user_value = 0
        self._user_minimum = 0
        self._user_maximum = 100
        self._label_template = "$value / $maximum - $percentage %"

        # Color settings
        self._background_color = QColor(30, 30, 30)
        self._progress_color = accent_colors.highlight  # QColor(210, 55, 130)

        self._completed_color = accent_colors.success
        self._border_color = QColor(50, 50, 50)

        # layout settings
        self._value_animation = QPropertyAnimation(self, b"_progressbar_value")
        self._value_animation.setDuration(200)
        self._value_animation.setEasingCurve(QEasingCurve.Type.OutCubic)

        # label on top of the progress bar
        self.center_label = QLabel(self)
        self.center_label.setAlignment(Qt.AlignCenter)
        self.center_label.setStyleSheet("color: white;")
        self.center_label.setMinimumSize(0, 0)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.center_label)
        self.setLayout(layout)

        self.update()

    @Property(str, doc="The template for the center label. Use $value, $maximum, and $percentage.")
    def label_template(self):
        """
        The template for the center label. Use $value, $maximum, and $percentage to insert the values.

        Examples:
        >>> progressbar.label_template = "$value / $maximum - $percentage %"
        >>> progressbar.label_template = "$value / $percentage %"

        """
        return self._label_template

    @label_template.setter
    def label_template(self, template):
        self._label_template = template
        self.set_value(self._user_value)
        self.update()

    @Property(float, designable=False)
    def _progressbar_value(self):
        """
        The current value of the progress bar.
        """
        return self._value

    @_progressbar_value.setter
    def _progressbar_value(self, val):
        self._value = val
        self.update()

    def _update_template(self):
        template = Template(self._label_template)
        return template.safe_substitute(
            value=self._user_value,
            maximum=self._user_maximum,
            percentage=int((self.map_value(self._user_value) / self._maximum) * 100),
        )

    @Slot(float)
    @Slot(int)
    def set_value(self, value):
        """
        Set the value of the progress bar.

        Args:
            value (float): The value to set.
        """
        if value > self._user_maximum:
            value = self._user_maximum
        elif value < self._user_minimum:
            value = self._user_minimum
        self._target_value = self.map_value(value)
        self._user_value = value
        self.center_label.setText(self._update_template())
        self.animate_progress()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        rect = self.rect().adjusted(10, 0, -10, -1)

        # Draw background
        painter.setBrush(self._background_color)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(rect, 10, 10)  # Rounded corners

        # Draw border
        painter.setBrush(Qt.NoBrush)
        painter.setPen(self._border_color)
        painter.drawRoundedRect(rect, 10, 10)

        # Determine progress color based on completion
        if self._value >= self._maximum:
            current_color = self._completed_color
        else:
            current_color = self._progress_color

        # Set clipping region to preserve the background's rounded corners
        progress_rect = rect.adjusted(
            0, 0, int(-rect.width() + (self._value / self._maximum) * rect.width()), 0
        )
        clip_path = QPainterPath()
        clip_path.addRoundedRect(QRectF(rect), 10, 10)  # Clip to the background's rounded corners
        painter.setClipPath(clip_path)

        # Draw progress bar
        painter.setBrush(current_color)
        painter.drawRect(progress_rect)  # Less rounded, no additional rounding

        painter.end()

    def animate_progress(self):
        """
        Animate the progress bar from the current value to the target value.
        """
        self._value_animation.stop()
        self._value_animation.setStartValue(self._value)
        self._value_animation.setEndValue(self._target_value)
        self._value_animation.start()

    @Property(float)
    def maximum(self):
        """
        The maximum value of the progress bar.
        """
        return self._user_maximum

    @maximum.setter
    def maximum(self, maximum: float):
        """
        Set the maximum value of the progress bar.
        """
        self.set_maximum(maximum)

    @Property(float)
    def minimum(self):
        """
        The minimum value of the progress bar.
        """
        return self._user_minimum

    @minimum.setter
    def minimum(self, minimum: float):
        self.set_minimum(minimum)

    @Property(float)
    def initial_value(self):
        """
        The initial value of the progress bar.
        """
        return self._user_value

    @initial_value.setter
    def initial_value(self, value: float):
        self.set_value(value)

    @Slot(float)
    def set_maximum(self, maximum: float):
        """
        Set the maximum value of the progress bar.

        Args:
            maximum (float): The maximum value.
        """
        self._user_maximum = maximum
        self.set_value(self._user_value)  # Update the value to fit the new range
        self.update()

    @Slot(float)
    def set_minimum(self, minimum: float):
        """
        Set the minimum value of the progress bar.

        Args:
            minimum (float): The minimum value.
        """
        self._user_minimum = minimum
        self.set_value(self._user_value)  # Update the value to fit the new range
        self.update()

    def map_value(self, value: float):
        """
        Map the user value to the range [0, 100*self._oversampling_factor] for the progress
        """
        return (
            (value - self._user_minimum) / (self._user_maximum - self._user_minimum) * self._maximum
        )

    def _get_label(self) -> str:
        """Return the label text. mostly used for testing rpc."""
        return self.center_label.text()


if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)

    progressBar = BECProgressBar()
    progressBar.show()
    progressBar.set_minimum(-100)
    progressBar.set_maximum(0)

    # Example of setting values
    def update_progress():
        value = progressBar._user_value + 2.5
        if value > progressBar._user_maximum:
            value = -100  # progressBar._maximum / progressBar._upsampling_factor
        progressBar.set_value(value)

    timer = QTimer()
    timer.timeout.connect(update_progress)
    timer.start(200)  # Update every half second

    sys.exit(app.exec())



import os
import re
from typing import Optional

from bec_lib.callback_handler import EventType
from pyqtgraph import SignalProxy
from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QListWidgetItem, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.ui_loader import UILoader
from bec_widgets.widgets.services.device_browser.device_item import DeviceItem


class DeviceBrowser(BECWidget, QWidget):
    """
    DeviceBrowser is a widget that displays all available devices in the current BEC session.
    """

    device_update: Signal = Signal()
    PLUGIN = True
    ICON_NAME = "lists"

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        config=None,
        client=None,
        gui_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.get_bec_shortcuts()
        self.ui = None
        self.ini_ui()

        self.proxy_device_update = SignalProxy(
            self.ui.filter_input.textChanged, rateLimit=500, slot=self.update_device_list
        )
        self.bec_dispatcher.client.callbacks.register(
            EventType.DEVICE_UPDATE, self.on_device_update
        )
        self.device_update.connect(self.update_device_list)

        self.update_device_list()

    def ini_ui(self) -> None:
        """
        Initialize the UI by loading the UI file and setting the layout.
        """
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        ui_file_path = os.path.join(os.path.dirname(__file__), "device_browser.ui")
        self.ui = UILoader(self).loader(ui_file_path)
        layout.addWidget(self.ui)
        self.setLayout(layout)

    def on_device_update(self, action: str, content: dict) -> None:
        """
        Callback for device update events. Triggers the device_update signal.

        Args:
            action (str): The action that triggered the event.
            content (dict): The content of the config update.
        """
        if action in ["add", "remove", "reload"]:
            self.device_update.emit()

    @Slot()
    def update_device_list(self) -> None:
        """
        Update the device list based on the filter input.
        There are two ways to trigger this function:
        1. By changing the text in the filter input.
        2. By emitting the device_update signal.

        Either way, the function will filter the devices based on the filter input text and update the device list.
        """
        filter_text = self.ui.filter_input.text()
        try:
            regex = re.compile(filter_text, re.IGNORECASE)
        except re.error:
            regex = None  # Invalid regex, disable filtering

        dev_list = self.ui.device_list
        dev_list.clear()
        for device in self.dev:
            if regex is None or regex.search(device):
                item = QListWidgetItem(dev_list)
                device_item = DeviceItem(device)

                # pylint: disable=protected-access
                tooltip = self.dev[device]._config.get("description", "")
                device_item.setToolTip(tooltip)
                item.setSizeHint(device_item.sizeHint())
                dev_list.setItemWidget(item, device_item)
                dev_list.addItem(item)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.utils.colors import apply_theme

    app = QApplication(sys.argv)
    apply_theme("light")
    widget = DeviceBrowser()
    widget.show()
    sys.exit(app.exec_())



"""Module for a thin wrapper (LinearRegionWrapper) around the LinearRegionItem in pyqtgraph.
The class is mainly designed for usage with the BECWaveform and 1D plots."""

from __future__ import annotations

import pyqtgraph as pg
from qtpy.QtCore import QObject, Signal, Slot
from qtpy.QtGui import QColor


class LinearRegionWrapper(QObject):
    """Wrapper class for the LinearRegionItem in pyqtgraph for 1D plots (BECWaveform)

    Args:
        plot_item (pg.PlotItem): The plot item to add the region selector to.
        parent (QObject): The parent object.
        color (QColor): The color of the region selector.
        hover_color (QColor): The color of the region selector when the mouse is over it.
    """

    # Signal with the region tuble (start, end)
    region_changed = Signal(tuple)

    def __init__(
        self, plot_item: pg.PlotItem, color: QColor = None, hover_color: QColor = None, parent=None
    ):
        super().__init__(parent)
        self.is_log_x = None
        self._edge_width = 2
        self.plot_item = plot_item
        self.linear_region_selector = pg.LinearRegionItem()
        self.proxy = None
        self.change_roi_color((color, hover_color))
        self.plot_item.ctrl.logXCheck.checkStateChanged.connect(self.check_log)
        self.plot_item.ctrl.logYCheck.checkStateChanged.connect(self.check_log)

    # Slot for changing the color of the region selector (edge and fill)
    @Slot(tuple)
    def change_roi_color(self, colors: tuple[QColor | str | tuple, QColor | str | tuple]):
        """Change the color and hover color of the region selector.
        Hover color means the color when the mouse is over the region.

        Args:
            colors (tuple): Tuple with the color and hover color
        """
        color, hover_color = colors
        if color is not None:
            self.linear_region_selector.setBrush(pg.mkBrush(color))
        if hover_color is not None:
            self.linear_region_selector.setHoverBrush(pg.mkBrush(hover_color))

    @Slot()
    def add_region_selector(self):
        """Add the region selector to the plot item"""
        self.plot_item.addItem(self.linear_region_selector)
        # Use proxy to limit the update rate of the region change signal to 10Hz
        self.proxy = pg.SignalProxy(
            self.linear_region_selector.sigRegionChanged,
            rateLimit=10,
            slot=self._region_change_proxy,
        )

    @Slot()
    def remove_region_selector(self):
        """Remove the region selector from the plot item"""
        self.proxy.disconnect()
        self.proxy = None
        self.plot_item.removeItem(self.linear_region_selector)

    def _region_change_proxy(self):
        """Emit the region change signal. If the plot is in log mode, convert the region to log."""
        x_low, x_high = self.linear_region_selector.getRegion()
        if self.is_log_x:
            x_low = 10**x_low
            x_high = 10**x_high
        self.region_changed.emit((x_low, x_high))

    @Slot()
    def check_log(self):
        """Check if the plot is in log mode."""
        self.is_log_x = self.plot_item.ctrl.logXCheck.isChecked()
        self.is_log_y = self.plot_item.ctrl.logYCheck.isChecked()

    def cleanup(self):
        """Cleanup the widget."""
        self.remove_region_selector()



from bec_lib.serialization import MsgpackSerialization
from bec_lib.utils import lazy_import_from
from qtpy.QtCore import QEventLoop, QSocketNotifier, QTimer

MessageObject = lazy_import_from("bec_lib.connector", ("MessageObject",))


class QtRedisMessageWaiter:
    def __init__(self, redis_connector, message_to_wait):
        self.ev_loop = QEventLoop()
        self.response = None
        self.connector = redis_connector
        self.message_to_wait = message_to_wait
        self.pubsub = redis_connector._redis_conn.pubsub()
        self.pubsub.subscribe(self.message_to_wait.endpoint)
        fd = self.pubsub.connection._sock.fileno()
        self.notifier = QSocketNotifier(fd, QSocketNotifier.Read)
        self.notifier.activated.connect(self._pubsub_readable)

    def _msg_received(self, msg_obj):
        self.response = msg_obj.value
        self.ev_loop.quit()

    def wait(self, timeout=1):
        timer = QTimer()
        timer.singleShot(timeout * 1000, self.ev_loop.quit)
        self.ev_loop.exec_()
        timer.stop()
        self.notifier.setEnabled(False)
        self.pubsub.close()
        return self.response

    def _pubsub_readable(self, fd):
        while True:
            msg = self.pubsub.get_message()
            if msg:
                if msg["type"] == "subscribe":
                    # get_message buffers, so we may already have the answer
                    # let's check...
                    continue
                else:
                    break
            else:
                return
        channel = msg["channel"].decode()
        msg = MessageObject(topic=channel, value=MsgpackSerialization.loads(msg["data"]))
        self.connector._execute_callback(self._msg_received, msg, {})



from __future__ import annotations

from typing import Literal

import pyqtgraph as pg
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QHBoxLayout, QWidget

from bec_widgets.utils.error_popups import SafeSlot


class ColorButton(QWidget):
    """
    A ColorButton that opens a dialog to select a color. Inherits from pyqtgraph.ColorButton.
    Patches event loop of the ColorDialog, if opened in another QDialog.
    """

    color_selected = Signal(str)

    PLUGIN = True
    ICON_NAME = "colors"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        self.button = pg.ColorButton()
        self.button.setFlat(True)
        self.button.clicked.connect(self.select_color)
        self.layout.addWidget(self.button)

    @SafeSlot()
    def select_color(self):
        self.origColor = self.button.color()
        self.button.colorDialog.setCurrentColor(self.button.color())
        self.button.colorDialog.open()
        self.button.colorDialog.exec()
        self.color_selected.emit(self.button.color().name())

    @SafeSlot(str)
    def set_color(self, color: tuple | str):
        """
        Set the color of the button.

        Args:
            color(tuple|str): The color to set.
        """
        self.button.setColor(color)

    def get_color(self, format: Literal["RGBA", "HEX"] = "RGBA") -> tuple | str:
        """
        Get the color of the button in the specified format.

        Args:
            format(Literal["RGBA", "HEX"]): The format of the returned color.

        Returns:
            tuple|str: The color in the specified format.
        """
        if format == "RGBA":
            return self.button.color().getRgb()
        if format == "HEX":
            return self.button.color().name()

    def cleanup(self):
        """
        Clean up the ColorButton.
        """
        self.button.colorDialog.close()
        self.button.colorDialog.deleteLater()



"""Module for a text box widget that displays text in plain and HTML format and adheres to the BECWidget interface & style."""

import re
from html.parser import HTMLParser

from bec_lib.logger import bec_logger
from pydantic import Field
from qtpy.QtWidgets import QTextEdit, QVBoxLayout, QWidget

from bec_widgets.utils.bec_connector import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot

logger = bec_logger.logger

DEFAULT_TEXT = "<h1>Welcome to the BEC Widget TextBox</h1><p>A widget that allows user to display text in plain and HTML format.</p><p>This is an example of displaying HTML text.</p>"


class TextBoxConfig(ConnectionConfig):
    """Configuration for the TextBox widget.

    Args:
        text (str, optional): The text to display in the widget. Defaults to None.
        is_html (bool, optional): Whether the text is in HTML format or not. Defaults to False.
    """

    text: str | None = Field(None, description="The text to display in the widget.")
    is_html: bool = Field(False, description="Whether the text is in HTML format or not.")


class TextBox(BECWidget, QWidget):
    """A widget that displays text in plain and HTML format

    Args:
        parent (QWidget, optional): The parent widget. Defaults to None.
        client ([type], optional): The client to use. Defaults to None.
        config ([type], optional): The config to use. Defaults to None.
        gui_id ([type], optional): The gui_id to use. Defaults to None.
    """

    PLUGIN = True
    USER_ACCESS = ["set_plain_text", "set_html_text"]
    ICON_NAME = "chat"

    def __init__(self, parent=None, client=None, config=None, gui_id=None, **kwargs):
        if config is None:
            config = TextBoxConfig(widget_class=self.__class__.__name__)
        else:
            if isinstance(config, dict):
                config = TextBoxConfig(**config)
            self.config = config
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)
        self.layout = QVBoxLayout(self)
        self.text_box_text_edit = QTextEdit(parent=self)
        self.layout.addWidget(self.text_box_text_edit)
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.config = config
        self.text_box_text_edit.setReadOnly(True)
        if self.config.text is not None:
            if self.config.is_html:
                self.set_html_text(self.config.text)
            else:
                self.set_plain_text(self.config.text)
        else:
            self.set_html_text(DEFAULT_TEXT)

    @SafeSlot(str)
    def set_plain_text(self, text: str) -> None:
        """Set the plain text of the widget.

        Args:
            text (str): The text to set.
        """
        self.text_box_text_edit.setPlainText(text)
        self.config.text = text
        self.config.is_html = False

    @SafeSlot(str)
    def set_html_text(self, text: str) -> None:
        """Set the HTML text of the widget.

        Args:
            text (str): The text to set.
        """
        self.text_box_text_edit.setHtml(text)
        self.config.text = text
        self.config.is_html = True

    @SafeProperty(str)
    def plain_text(self) -> str:
        """Get the text of the widget.

        Returns:
            str: The text of the widget.
        """
        return self.text_box_text_edit.toPlainText()

    @plain_text.setter
    def plain_text(self, text: str) -> None:
        """Set the text of the widget.

        Args:
            text (str): The text to set.
        """
        self.set_plain_text(text)

    @SafeProperty(str)
    def html_text(self) -> str:
        """Get the HTML text of the widget.

        Returns:
            str: The HTML text of the widget.
        """
        return self.text_box_text_edit.toHtml()

    @html_text.setter
    def html_text(self, text: str) -> None:
        """Set the HTML text of the widget.

        Args:
            text (str): The HTML text to set.
        """
        self.set_html_text(text)


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = TextBox()
    widget.show()
    sys.exit(app.exec())



import importlib.metadata
import json
import os
import site
import sys
import sysconfig
from pathlib import Path

from bec_qthemes import material_icon
from qtpy import PYSIDE6
from qtpy.QtGui import QIcon

from bec_widgets.utils.bec_plugin_helper import user_widget_plugin

if PYSIDE6:
    from PySide6.scripts.pyside_tool import (
        _extend_path_var,
        init_virtual_env,
        is_pyenv_python,
        is_virtual_env,
        qt_tool_wrapper,
        ui_tool_binary,
    )

import bec_widgets


def designer_material_icon(icon_name: str) -> QIcon:
    """
    Create a QIcon for the BECDesigner with the given material icon name.

    Args:
        icon_name (str): The name of the material icon.

    Returns:
        QIcon: The QIcon for the material icon.
    """
    return QIcon(material_icon(icon_name, filled=True, convert_to_pixmap=True))


def list_editable_packages() -> set[str]:
    """
    List all editable packages in the environment.

    Returns:
        set: A set of paths to editable packages.
    """

    editable_packages = set()

    # Get site-packages directories
    site_packages = site.getsitepackages()
    if hasattr(site, "getusersitepackages"):
        site_packages.append(site.getusersitepackages())

    for dist in importlib.metadata.distributions():
        location = dist.locate_file("").resolve()
        is_editable = all(not str(location).startswith(site_pkg) for site_pkg in site_packages)

        if is_editable:
            editable_packages.add(str(location))

    for packages in site_packages:
        # all dist-info directories in site-packages that contain a direct_url.json file
        dist_info_dirs = Path(packages).rglob("*.dist-info")
        for dist_info_dir in dist_info_dirs:
            direct_url = dist_info_dir / "direct_url.json"
            if not direct_url.exists():
                continue
            # load the json file and get the path to the package
            with open(direct_url, "r", encoding="utf-8") as f:
                data = json.load(f)
                path = data.get("url", "")
                if path.startswith("file://"):
                    path = path[7:]
                    editable_packages.add(path)

    return editable_packages


def patch_designer(cmd_args: list[str] = []):  # pragma: no cover
    if not PYSIDE6:
        print("PYSIDE6 is not available in the environment. Cannot patch designer.")
        return

    init_virtual_env()

    major_version = sys.version_info[0]
    minor_version = sys.version_info[1]
    os.environ["PY_MAJOR_VERSION"] = str(major_version)
    os.environ["PY_MINOR_VERSION"] = str(minor_version)

    if sys.platform == "win32":
        if is_virtual_env():
            _extend_path_var("PATH", os.fspath(Path(sys._base_executable).parent), True)
    else:
        if sys.platform == "linux":
            env_var = "LD_PRELOAD"
            current_pid = os.getpid()
            with open(f"/proc/{current_pid}/maps", "rt") as f:
                for line in f:
                    if "libpython" in line:
                        lib_path = line.split()[-1]
                        os.environ[env_var] = lib_path
                        break

        elif sys.platform == "darwin":
            suffix = ".dylib"
            env_var = "DYLD_INSERT_LIBRARIES"
            version = f"{major_version}.{minor_version}"
            library_name = f"libpython{version}{suffix}"
            lib_path = str(Path(sysconfig.get_config_var("LIBDIR")) / library_name)
            os.environ[env_var] = lib_path
        else:
            raise RuntimeError(f"Unsupported platform: {sys.platform}")

        if is_pyenv_python() or is_virtual_env():
            # append all editable packages to the PYTHONPATH
            editable_packages = list_editable_packages()
            for pckg in editable_packages:
                _extend_path_var("PYTHONPATH", pckg, True)
    qt_tool_wrapper(ui_tool_binary("designer"), cmd_args)


def find_plugin_paths(base_path: Path):
    """
    Recursively find all directories containing a .pyproject file.
    """
    plugin_paths = []
    for path in base_path.rglob("*.pyproject"):
        plugin_paths.append(str(path.parent))
    return plugin_paths


def set_plugin_environment_variable(plugin_paths):
    """
    Set the PYSIDE_DESIGNER_PLUGINS environment variable with the given plugin paths.
    """
    current_paths = os.environ.get("PYSIDE_DESIGNER_PLUGINS", "")
    if current_paths:
        current_paths = current_paths.split(os.pathsep)
    else:
        current_paths = []

    current_paths.extend(plugin_paths)
    os.environ["PYSIDE_DESIGNER_PLUGINS"] = os.pathsep.join(current_paths)


# Patch the designer function
def open_designer(cmd_args: list[str] = []):  # pragma: no cover
    if not PYSIDE6:
        print("PYSIDE6 is not available in the environment. Exiting...")
        return
    base_dir = Path(os.path.dirname(bec_widgets.__file__)).resolve()

    plugin_paths = find_plugin_paths(base_dir)
    if (plugin_repo := user_widget_plugin()) and isinstance(plugin_repo.__file__, str):
        plugin_repo_dir = Path(os.path.dirname(plugin_repo.__file__)).resolve()
        plugin_paths.extend(find_plugin_paths(plugin_repo_dir))

    set_plugin_environment_variable(plugin_paths)

    patch_designer(cmd_args)


def main():
    open_designer(sys.argv[1:])


if __name__ == "__main__":  # pragma: no cover
    main()



import numpy as np
from qtpy.QtCore import Property, QSize, Qt, Slot
from qtpy.QtGui import QBrush, QColor, QPainter, QPainterPath, QPen
from qtpy.QtWidgets import QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors, get_theme_palette


class PositionIndicator(BECWidget, QWidget):
    """
    Display a position within a defined range, e.g. motor limits.
    """

    USER_ACCESS = [
        "set_value",
        "set_range",
        "vertical",
        "vertical.setter",
        "indicator_width",
        "indicator_width.setter",
        "rounded_corners",
        "rounded_corners.setter",
    ]
    PLUGIN = True
    ICON_NAME = "horizontal_distribute"

    def __init__(self, parent=None, client=None, config=None, gui_id=None, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.position = 50
        self.min_value = 0
        self.max_value = 100
        self.scaling_factor = 0.5
        self.is_vertical = False
        self._current_indicator_position = 0
        self._draw_position = 0
        self._rounded_corners = 10
        self._indicator_width = 2
        self._indicator_color = get_accent_colors().success
        self._background_color = get_theme_palette().mid().color()
        self._use_color_palette = True

    def set_range(self, min_value: float, max_value: float):
        """
        Set the range of the position indicator

        Args:
            min_value(float): Minimum value of the range
            max_value(float): Maximum value of the range
        """
        self.minimum = min_value
        self.maximum = max_value

    @Property(float)
    def minimum(self):
        """
        Property to get the minimum value of the position indicator
        """
        return self.min_value

    @minimum.setter
    def minimum(self, min_value: float):
        """
        Setter for the minimum property

        Args:
            min_value: The minimum value of the position indicator
        """
        self.min_value = min_value
        self.update()

    @Property(float)
    def maximum(self):
        """
        Property to get the maximum value of the position indicator
        """
        return self.max_value

    @maximum.setter
    def maximum(self, max_value: float):
        """
        Setter for the maximum property

        Args:
            max_value: The maximum value of the position indicator
        """
        self.max_value = max_value
        self.update()

    @Property(bool)
    def vertical(self):
        """
        Property to determine the orientation of the position indicator
        """
        return self.is_vertical

    @vertical.setter
    def vertical(self, is_vertical: bool):
        """
        Setter for the vertical property

        Args:
            is_vertical: True if the indicator should be vertical, False if horizontal
        """

        self.is_vertical = is_vertical
        self.update()

    @Property(float)
    def value(self):
        """
        Property to get the current value of the position indicator
        """
        return self.position

    @value.setter
    def value(self, position: float):
        """
        Setter for the value property

        Args:
            position: The new position of the indicator
        """
        self.set_value(position)

    @Property(int)
    def indicator_width(self):
        """
        Property to get the width of the indicator
        """
        return self._indicator_width

    @indicator_width.setter
    def indicator_width(self, width: int):
        """
        Setter for the indicator width property

        Args:
            width: The new width of the indicator
        """
        self._indicator_width = width
        self.update()

    @Property(int)
    def rounded_corners(self):
        """
        Property to get the rounded corners of the position indicator
        """
        return self._rounded_corners

    @rounded_corners.setter
    def rounded_corners(self, value: int):
        """
        Setter for the rounded corners property

        Args:
            value: The new value for the rounded corners
        """
        self._rounded_corners = value
        self.update()

    @Property(QColor)
    def indicator_color(self):
        """
        Property to get the color of the indicator
        """
        return self._indicator_color

    @indicator_color.setter
    def indicator_color(self, color: QColor):
        """
        Setter for the indicator color property

        Args:
            color: The new color for the indicator
        """
        self._indicator_color = color
        self.update()

    @Property(QColor)
    def background_color(self):
        """
        Property to get the background color of the position indicator
        """
        return self._background_color

    @background_color.setter
    def background_color(self, color: QColor):
        """
        Setter for the background color property

        Args:
            color: The new background color
        """
        self._background_color = color
        self.update()

    @Property(bool)
    def use_color_palette(self):
        """
        Property to determine if the indicator should use the color palette or the custom color.
        """
        return self._use_color_palette

    @use_color_palette.setter
    def use_color_palette(self, use_palette: bool):
        """
        Setter for the use color palette property

        Args:
            use_palette: True if the indicator should use the color palette, False if custom color
        """
        self._use_color_palette = use_palette
        self.update()

    # @Property(float)
    @Slot(int)
    @Slot(float)
    def set_value(self, position: float):
        """
        Set the position of the indicator

        Args:
            position: The new position of the indicator
        """
        self.position = position
        self.update()

    def _get_indicator_color(self):
        if self._use_color_palette:
            return get_accent_colors().success
        return self._indicator_color

    def _get_background_brush(self):
        if self._use_color_palette:
            return get_theme_palette().mid()
        return QBrush(self._background_color)

    def paintEvent(self, event):
        painter = QPainter(self)
        width = self.width()
        height = self.height()

        # Set up the brush for the background
        painter.setBrush(self._get_background_brush())

        # Create a QPainterPath with a rounded rectangle for clipping
        path = QPainterPath()
        path.addRoundedRect(0, 0, width, height, self._rounded_corners, self._rounded_corners)

        # Set clipping to the rounded rectangle
        painter.setClipPath(path)

        # Draw the rounded rectangle background first
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(0, 0, width, height, self._rounded_corners, self._rounded_corners)

        # get the position scaled to the defined min and max values
        self._current_indicator_position = position = np.interp(
            self.position, [self.min_value, self.max_value], [0, 100]
        )

        if self.is_vertical:
            # If vertical, rotate the coordinate system by -90 degrees
            painter.translate(width // 2, height // 2)  # Move origin to center
            painter.rotate(-90)  # Rotate by -90 degrees for vertical drawing
            painter.translate(-height // 2, -width // 2)  # Restore the origin for drawing

            # Switch width and height for the vertical orientation
            width, height = height, width

        # Draw the moving vertical indicator, respecting the clip path
        self._draw_position = x_pos = round(
            position * width / 100
        )  # Position for the vertical line

        indicator_pen = QPen(self._get_indicator_color(), self._indicator_width)
        painter.setPen(indicator_pen)
        painter.drawLine(x_pos, 0, x_pos, height)

        painter.end()

    def minimumSizeHint(self):
        # Set the smallest possible size
        return QSize(10, 10)


if __name__ == "__main__":  # pragma: no cover
    from bec_qthemes import setup_theme
    from qtpy.QtWidgets import QApplication, QSlider, QVBoxLayout

    app = QApplication([])
    setup_theme("dark")
    # Create position indicator and slider
    position_indicator = PositionIndicator()
    # position_indicator.set_range(0, 1)
    slider = QSlider(Qt.Horizontal)
    slider.valueChanged.connect(lambda value: position_indicator.set_value(value))
    position_indicator.is_vertical = False
    # position_indicator.set_value(100)
    layout = QVBoxLayout()
    layout.addWidget(position_indicator)
    layout.addWidget(slider)

    widget = QWidget()
    widget.setLayout(layout)
    widget.show()

    app.exec_()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QScrollArea,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors, get_theme_palette
from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton


class PaletteViewer(BECWidget, QWidget):
    """
    This class is a widget that displays current palette colors.
    """

    ICON_NAME = "palette"
    RPC = False

    def __init__(self, *args, parent=None, **kwargs):
        super().__init__(parent=parent, theme_update=True, **kwargs)
        self.setFixedSize(400, 600)
        layout = QVBoxLayout(self)
        dark_mode_button = DarkModeButton(self)
        layout.addWidget(dark_mode_button)

        # Create a scroll area to hold the color boxes
        scroll_area = QScrollArea(self)
        scroll_area.setWidgetResizable(True)
        scroll_area.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # Create a frame to hold the color boxes
        self.frame = QFrame(self)
        self.frame_layout = QGridLayout(self.frame)
        self.frame_layout.setSpacing(0)
        self.frame_layout.setContentsMargins(0, 0, 0, 0)

        scroll_area.setWidget(self.frame)
        layout.addWidget(scroll_area)

        self.setLayout(layout)

        self.update_palette()

    def apply_theme(self, theme) -> None:
        """
        Apply the theme to the widget.

        Args:
            theme (str): The theme to apply.
        """
        self.update_palette()

    def clear_palette(self) -> None:
        """
        Clear the palette colors from the frame.
        Recursively removes all widgets and layouts in the frame layout.
        """
        # Iterate over all items in the layout in reverse to safely remove them
        for i in reversed(range(self.frame_layout.count())):
            item = self.frame_layout.itemAt(i)

            # If the item is a layout, clear its contents
            if isinstance(item, QHBoxLayout):
                # Recursively remove all widgets from the layout
                for j in reversed(range(item.count())):
                    widget = item.itemAt(j).widget()
                    if widget:
                        item.removeWidget(widget)
                        widget.deleteLater()
                self.frame_layout.removeItem(item)

            # If the item is a widget, remove and delete it
            elif item.widget():
                widget = item.widget()
                self.frame_layout.removeWidget(widget)
                widget.deleteLater()

    def update_palette(self) -> None:
        """
        Update the palette colors in the frame.
        """
        self.clear_palette()
        palette_label = QLabel("Palette Colors (e.g. palette.windowText().color())")
        palette_label.setStyleSheet("font-weight: bold;")
        self.frame_layout.addWidget(palette_label, 0, 0)

        palette = get_theme_palette()
        # Add the palette colors (roles) to the frame
        palette_roles = [
            palette.windowText,
            palette.toolTipText,
            palette.placeholderText,
            palette.text,
            palette.buttonText,
            palette.highlight,
            palette.link,
            palette.light,
            palette.midlight,
            palette.mid,
            palette.shadow,
            palette.button,
            palette.brightText,
            palette.toolTipBase,
            palette.alternateBase,
            palette.dark,
            palette.base,
            palette.window,
            palette.highlightedText,
            palette.linkVisited,
        ]

        offset = 1
        for i, pal in enumerate(palette_roles):
            i += offset
            color = pal().color()
            label_layout = QHBoxLayout()
            color_label = QLabel(f"{pal().color().name()} ({pal.__name__})")
            background_label = self.background_label_with_clipboard(color)
            label_layout.addWidget(color_label)
            label_layout.addWidget(background_label)
            self.frame_layout.addLayout(label_layout, i, 0)

        # add a horizontal spacer
        spacer = QLabel()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.frame_layout.addWidget(spacer, i + 1, 0)

        accent_colors_label = QLabel("Accent Colors (e.g. accent_colors.default)")
        accent_colors_label.setStyleSheet("font-weight: bold;")
        self.frame_layout.addWidget(accent_colors_label, i + 2, 0)

        accent_colors = get_accent_colors()
        items = [
            (accent_colors.default, "default"),
            (accent_colors.success, "success"),
            (accent_colors.warning, "warning"),
            (accent_colors.emergency, "emergency"),
            (accent_colors.highlight, "highlight"),
        ]

        offset = len(palette_roles) + 2
        for i, (color, name) in enumerate(items):
            i += offset
            label_layout = QHBoxLayout()
            color_label = QLabel(f"{color.name()} ({name})")
            background_label = self.background_label_with_clipboard(color)
            label_layout.addWidget(color_label)
            label_layout.addWidget(background_label)
            self.frame_layout.addLayout(label_layout, i + 2, 0)

    def background_label_with_clipboard(self, color) -> QLabel:
        """
        Create a label with a background color that copies the color to the clipboard when clicked.

        Args:
            color (QColor): The color to display in the background.

        Returns:
            QLabel: The label with the background color.
        """
        button = QLabel()
        button.setStyleSheet(f"QLabel {{ background-color: {color.name()}; }}")
        button.setToolTip("Click to copy color to clipboard")
        button.setCursor(Qt.PointingHandCursor)
        button.mousePressEvent = lambda event: QApplication.clipboard().setText(color.name())
        return button


if __name__ == "__main__":  # pragma: no cover
    import sys

    app = QApplication(sys.argv)
    viewer = PaletteViewer()
    viewer.show()
    sys.exit(app.exec_())



import os

from qtpy.QtWidgets import QFrame, QScrollArea, QVBoxLayout

from bec_widgets.utils import UILoader
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.settings_dialog import SettingWidget


class ScatterCurveSettings(SettingWidget):
    def __init__(self, parent=None, target_widget=None, popup=False, *args, **kwargs):
        super().__init__(parent=parent, *args, **kwargs)

        # This is a settings widget that depends on the target widget
        # and should mirror what is in the target widget.
        # Saving settings for this widget could result in recursively setting the target widget.
        self.setProperty("skip_settings", True)

        current_path = os.path.dirname(__file__)
        if popup:
            form = UILoader().load_ui(
                os.path.join(current_path, "scatter_curve_settings_horizontal.ui"), self
            )
        else:
            form = UILoader().load_ui(
                os.path.join(current_path, "scatter_curve_settings_vertical.ui"), self
            )

        self.target_widget = target_widget
        self.popup = popup

        # # Scroll area
        self.scroll_area = QScrollArea(self)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setWidget(form)

        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.scroll_area)
        self.ui = form

        self.fetch_all_properties()

        self.target_widget.scatter_waveform_property_changed.connect(self.fetch_all_properties)
        if popup is False:
            self.ui.button_apply.clicked.connect(self.accept_changes)

    @SafeSlot()
    def fetch_all_properties(self):
        """
        Fetch all properties from the target widget and update the settings widget.
        """
        if not self.target_widget:
            return

        # Get properties from the target widget
        color_map = getattr(self.target_widget, "color_map", None)

        # Default values for device properties
        x_name, x_entry = None, None
        y_name, y_entry = None, None
        z_name, z_entry = None, None

        # Safely access device properties
        if hasattr(self.target_widget, "main_curve") and self.target_widget.main_curve:
            if hasattr(self.target_widget.main_curve, "config"):
                config = self.target_widget.main_curve.config

                if hasattr(config, "x_device") and config.x_device:
                    x_name = getattr(config.x_device, "name", None)
                    x_entry = getattr(config.x_device, "entry", None)

                if hasattr(config, "y_device") and config.y_device:
                    y_name = getattr(config.y_device, "name", None)
                    y_entry = getattr(config.y_device, "entry", None)

                if hasattr(config, "z_device") and config.z_device:
                    z_name = getattr(config.z_device, "name", None)
                    z_entry = getattr(config.z_device, "entry", None)

        # Apply the properties to the settings widget
        if hasattr(self.ui, "color_map"):
            self.ui.color_map.colormap = color_map

        if hasattr(self.ui, "x_name"):
            self.ui.x_name.set_device(x_name)
        if hasattr(self.ui, "x_entry") and x_entry is not None:
            self.ui.x_entry.setText(x_entry)

        if hasattr(self.ui, "y_name"):
            self.ui.y_name.set_device(y_name)
        if hasattr(self.ui, "y_entry") and y_entry is not None:
            self.ui.y_entry.setText(y_entry)

        if hasattr(self.ui, "z_name"):
            self.ui.z_name.set_device(z_name)
        if hasattr(self.ui, "z_entry") and z_entry is not None:
            self.ui.z_entry.setText(z_entry)

    @SafeSlot()
    def accept_changes(self):
        """
        Apply all properties from the settings widget to the target widget.
        """
        x_name = self.ui.x_name.text()
        x_entry = self.ui.x_entry.text()
        y_name = self.ui.y_name.text()
        y_entry = self.ui.y_entry.text()
        z_name = self.ui.z_name.text()
        z_entry = self.ui.z_entry.text()
        validate_bec = self.ui.validate_bec.checked
        color_map = self.ui.color_map.colormap

        self.target_widget.plot(
            x_name=x_name,
            y_name=y_name,
            z_name=z_name,
            x_entry=x_entry,
            y_entry=y_entry,
            z_entry=z_entry,
            color_map=color_map,
            validate_bec=validate_bec,
        )

    def cleanup(self):
        self.ui.x_name.close()
        self.ui.x_name.deleteLater()
        self.ui.x_entry.close()
        self.ui.x_entry.deleteLater()
        self.ui.y_name.close()
        self.ui.y_name.deleteLater()
        self.ui.y_entry.close()
        self.ui.y_entry.deleteLater()
        self.ui.z_name.close()
        self.ui.z_name.deleteLater()
        self.ui.z_entry.close()
        self.ui.z_entry.deleteLater()



import pyqtgraph as pg

from bec_widgets.utils.round_frame import RoundedFrame
from bec_widgets.widgets.plots.plot_base import BECViewBox


class ImageROIPlot(RoundedFrame):
    """
    A widget for displaying an image with a region of interest (ROI) overlay.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.content_widget = pg.GraphicsLayoutWidget(self)
        self.layout.addWidget(self.content_widget)
        self.plot_item = pg.PlotItem(viewBox=BECViewBox(enableMenu=True))
        self.content_widget.addItem(self.plot_item)
        self.curve_color = "w"

        self.apply_plot_widget_style()

    def apply_theme(self, theme: str):
        if theme == "dark":
            self.curve_color = "w"
        else:
            self.curve_color = "k"
        for curve in self.plot_item.curves:
            curve.setPen(pg.mkPen(self.curve_color, width=3))
        super().apply_theme(theme)

    def cleanup_pyqtgraph(self):
        """Cleanup pyqtgraph items."""
        self.plot_item.vb.menu.close()
        self.plot_item.vb.menu.deleteLater()
        self.plot_item.ctrlMenu.close()
        self.plot_item.ctrlMenu.deleteLater()



import threading


class ThreadTracker:
    def __init__(self, exclude_names=None):
        self.exclude_names = exclude_names if exclude_names else []
        self.initial_threads = self._capture_threads()

    def _capture_threads(self):
        return set(
            th
            for th in threading.enumerate()
            if not any(ex_name in th.name for ex_name in self.exclude_names)
            and th is not threading.main_thread()
        )

    def _thread_info(self, threads):
        return ", \n".join(f"{th.name}(ID: {th.ident})" for th in threads)

    def check_unfinished_threads(self):
        current_threads = self._capture_threads()
        additional_threads = current_threads - self.initial_threads
        closed_threads = self.initial_threads - current_threads
        if additional_threads:
            raise Exception(
                f"###### Initial threads ######:\n {self._thread_info(self.initial_threads)}\n"
                f"###### Current threads ######:\n {self._thread_info(current_threads)}\n"
                f"###### Closed threads ######:\n {self._thread_info(closed_threads)}\n"
                f"###### Unfinished threads ######:\n {self._thread_info(additional_threads)}"
            )
        else:
            print(
                "All threads properly closed.\n"
                f"###### Initial threads ######:\n {self._thread_info(self.initial_threads)}\n"
                f"###### Current threads ######:\n {self._thread_info(current_threads)}\n"
                f"###### Closed threads ######:\n {self._thread_info(closed_threads)}"
            )



"""Module for a PositionerBox2D widget to control two positioner devices."""

from __future__ import annotations

import os
from typing import Literal

from bec_lib.device import Positioner
from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from qtpy.QtCore import Signal
from qtpy.QtGui import QDoubleValidator
from qtpy.QtWidgets import QDoubleSpinBox

from bec_widgets.utils import UILoader
from bec_widgets.utils.colors import set_theme
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.widgets.control.device_control.positioner_box._base import PositionerBoxBase
from bec_widgets.widgets.control.device_control.positioner_box._base.positioner_box_base import (
    DeviceUpdateUIComponents,
)

logger = bec_logger.logger

MODULE_PATH = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))

DeviceId = Literal["horizontal", "vertical"]


class PositionerBox2D(PositionerBoxBase):
    """Simple Widget to control two positioners in box form"""

    ui_file = "positioner_box_2d.ui"

    PLUGIN = True
    RPC = True
    USER_ACCESS = ["set_positioner_hor", "set_positioner_ver"]

    device_changed_hor = Signal(str, str)
    device_changed_ver = Signal(str, str)
    # Signals emitted to inform listeners about a position update
    position_update_hor = Signal(float)
    position_update_ver = Signal(float)

    def __init__(
        self,
        parent=None,
        device_hor: Positioner | str | None = None,
        device_ver: Positioner | str | None = None,
        **kwargs,
    ):
        """Initialize the PositionerBox widget.

        Args:
            parent: The parent widget.
            device_hor (Positioner | str): The first device to control - assigned the horizontal axis.
            device_ver (Positioner | str): The second device to control - assigned the vertical axis.
        """
        super().__init__(parent=parent, **kwargs)

        self._device_hor = ""
        self._device_ver = ""
        self._limits_hor = None
        self._limits_ver = None
        self._dialog = None
        if self.current_path == "":
            self.current_path = os.path.dirname(__file__)
        self.init_ui()
        self.device_hor = device_hor
        self.device_ver = device_ver

        self.connect_ui()

    def init_ui(self):
        """Init the ui"""
        self.device_changed_hor.connect(self.on_device_change_hor)
        self.device_changed_ver.connect(self.on_device_change_ver)

        self.ui = UILoader(self).loader(os.path.join(self.current_path, self.ui_file))
        self.setpoint_validator_hor = QDoubleValidator()
        self.setpoint_validator_ver = QDoubleValidator()

    def connect_ui(self):
        """Connect the UI components to signals, data, or routines"""
        self.addWidget(self.ui)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)

        def _init_ui(val: QDoubleValidator, device_id: DeviceId):
            ui = self._device_ui_components_hv(device_id)
            tweak_inc = (
                self.on_tweak_inc_hor if device_id == "horizontal" else self.on_tweak_inc_ver
            )
            tweak_dec = (
                self.on_tweak_dec_hor if device_id == "horizontal" else self.on_tweak_dec_ver
            )
            ui["setpoint"].setValidator(val)
            ui["setpoint"].returnPressed.connect(
                self.on_setpoint_change_hor
                if device_id == "horizontal"
                else self.on_setpoint_change_ver
            )
            ui["stop"].setToolTip("Stop")
            ui["step_size"].setStepType(QDoubleSpinBox.StepType.AdaptiveDecimalStepType)
            ui["tweak_increase"].clicked.connect(tweak_inc)
            ui["tweak_decrease"].clicked.connect(tweak_dec)

        _init_ui(self.setpoint_validator_hor, "horizontal")
        _init_ui(self.setpoint_validator_ver, "vertical")

        self.ui.stop_button.button.clicked.connect(self.on_stop)

        self.ui.step_decrease_hor.clicked.connect(self.on_step_dec_hor)
        self.ui.step_decrease_ver.clicked.connect(self.on_step_dec_ver)
        self.ui.step_increase_hor.clicked.connect(self.on_step_inc_hor)
        self.ui.step_increase_ver.clicked.connect(self.on_step_inc_ver)

        self.ui.tool_button_hor.clicked.connect(
            self._open_dialog_selection(self.set_positioner_hor)
        )
        self.ui.tool_button_ver.clicked.connect(
            self._open_dialog_selection(self.set_positioner_ver)
        )
        icon = material_icon(icon_name="edit_note", size=(16, 16), convert_to_pixmap=False)
        self.ui.tool_button_hor.setIcon(icon)
        self.ui.tool_button_ver.setIcon(icon)

        step_tooltip = "Step by the step size"
        tweak_tooltip = "Tweak by 1/10th the step size"

        for b in [
            self.ui.step_increase_hor,
            self.ui.step_increase_ver,
            self.ui.step_decrease_hor,
            self.ui.step_decrease_ver,
        ]:
            b.setToolTip(step_tooltip)

        for b in [
            self.ui.tweak_increase_hor,
            self.ui.tweak_increase_ver,
            self.ui.tweak_decrease_hor,
            self.ui.tweak_decrease_ver,
        ]:
            b.setToolTip(tweak_tooltip)

        icon_options = {"size": (16, 16), "convert_to_pixmap": False}
        self.ui.tweak_increase_hor.setIcon(
            material_icon(icon_name="keyboard_arrow_right", **icon_options)
        )
        self.ui.step_increase_hor.setIcon(
            material_icon(icon_name="keyboard_double_arrow_right", **icon_options)
        )
        self.ui.tweak_decrease_hor.setIcon(
            material_icon(icon_name="keyboard_arrow_left", **icon_options)
        )
        self.ui.step_decrease_hor.setIcon(
            material_icon(icon_name="keyboard_double_arrow_left", **icon_options)
        )
        self.ui.tweak_increase_ver.setIcon(
            material_icon(icon_name="keyboard_arrow_up", **icon_options)
        )
        self.ui.step_increase_ver.setIcon(
            material_icon(icon_name="keyboard_double_arrow_up", **icon_options)
        )
        self.ui.tweak_decrease_ver.setIcon(
            material_icon(icon_name="keyboard_arrow_down", **icon_options)
        )
        self.ui.step_decrease_ver.setIcon(
            material_icon(icon_name="keyboard_double_arrow_down", **icon_options)
        )

    @SafeProperty(str)
    def device_hor(self):
        """SafeProperty to set the device"""
        return self._device_hor

    @device_hor.setter
    def device_hor(self, value: str):
        """Setter, checks if device is a string"""
        if not value or not isinstance(value, str):
            return
        if not self._check_device_is_valid(value):
            return
        if value == self.device_ver:
            return
        old_device = self._device_hor
        self._device_hor = value
        self.label = f"{self._device_hor}, {self._device_ver}"
        self.device_changed_hor.emit(old_device, value)
        self._init_device(self.device_hor, self.position_update_hor.emit, self.update_limits_hor)

    @SafeProperty(str)
    def device_ver(self):
        """SafeProperty to set the device"""
        return self._device_ver

    @device_ver.setter
    def device_ver(self, value: str):
        """Setter, checks if device is a string"""
        if not value or not isinstance(value, str):
            return
        if not self._check_device_is_valid(value):
            return
        if value == self.device_hor:
            return
        old_device = self._device_ver
        self._device_ver = value
        self.label = f"{self._device_hor}, {self._device_ver}"
        self.device_changed_ver.emit(old_device, value)
        self._init_device(self.device_ver, self.position_update_ver.emit, self.update_limits_ver)

    @SafeProperty(bool)
    def hide_device_selection(self):
        """Hide the device selection"""
        return not self.ui.tool_button_hor.isVisible()

    @hide_device_selection.setter
    def hide_device_selection(self, value: bool):
        """Set the device selection visibility"""
        self.ui.tool_button_hor.setVisible(not value)
        self.ui.tool_button_ver.setVisible(not value)

    @SafeProperty(bool)
    def hide_device_boxes(self):
        """Hide the device selection"""
        return not self.ui.device_box_hor.isVisible()

    @hide_device_boxes.setter
    def hide_device_boxes(self, value: bool):
        """Set the device selection visibility"""
        self.ui.device_box_hor.setVisible(not value)
        self.ui.device_box_ver.setVisible(not value)

    @SafeSlot(bool)
    def show_device_selection(self, value: bool):
        """Show the device selection

        Args:
            value (bool): Show the device selection
        """
        self.hide_device_selection = not value

    @SafeSlot(str)
    def set_positioner_hor(self, positioner: str | Positioner):
        """Set the device

        Args:
            positioner (Positioner | str) : Positioner to set, accepts str or the device
        """
        if isinstance(positioner, Positioner):
            positioner = positioner.name
        self.device_hor = positioner

    @SafeSlot(str)
    def set_positioner_ver(self, positioner: str | Positioner):
        """Set the device

        Args:
            positioner (Positioner | str) : Positioner to set, accepts str or the device
        """
        if isinstance(positioner, Positioner):
            positioner = positioner.name
        self.device_ver = positioner

    @SafeSlot(str, str)
    def on_device_change_hor(self, old_device: str, new_device: str):
        """Upon changing the device, a check will be performed if the device is a Positioner.

        Args:
            old_device (str): The old device name.
            new_device (str): The new device name.
        """
        if not self._check_device_is_valid(new_device):
            return
        self._on_device_change(
            old_device,
            new_device,
            self.position_update_hor.emit,
            self.update_limits_hor,
            self.on_device_readback_hor,
            self._device_ui_components_hv("horizontal"),
        )

    @SafeSlot(str, str)
    def on_device_change_ver(self, old_device: str, new_device: str):
        """Upon changing the device, a check will be performed if the device is a Positioner.

        Args:
            old_device (str): The old device name.
            new_device (str): The new device name.
        """
        if not self._check_device_is_valid(new_device):
            return
        self._on_device_change(
            old_device,
            new_device,
            self.position_update_ver.emit,
            self.update_limits_ver,
            self.on_device_readback_ver,
            self._device_ui_components_hv("vertical"),
        )

    def _device_ui_components_hv(self, device: DeviceId) -> DeviceUpdateUIComponents:
        if device == "horizontal":
            return {
                "spinner": self.ui.spinner_widget_hor,
                "position_indicator": self.ui.position_indicator_hor,
                "readback": self.ui.readback_hor,
                "setpoint": self.ui.setpoint_hor,
                "step_size": self.ui.step_size_hor,
                "device_box": self.ui.device_box_hor,
                "stop": self.ui.stop_button,
                "tweak_increase": self.ui.tweak_increase_hor,
                "tweak_decrease": self.ui.tweak_decrease_hor,
                "units": self.ui.units_hor,
            }
        elif device == "vertical":
            return {
                "spinner": self.ui.spinner_widget_ver,
                "position_indicator": self.ui.position_indicator_ver,
                "readback": self.ui.readback_ver,
                "setpoint": self.ui.setpoint_ver,
                "step_size": self.ui.step_size_ver,
                "device_box": self.ui.device_box_ver,
                "stop": self.ui.stop_button,
                "tweak_increase": self.ui.tweak_increase_ver,
                "tweak_decrease": self.ui.tweak_decrease_ver,
                "units": self.ui.units_ver,
            }
        else:
            raise ValueError(f"Device {device} is not represented by this UI")

    def _device_ui_components(self, device: str):
        if device == self.device_hor:
            return self._device_ui_components_hv("horizontal")
        if device == self.device_ver:
            return self._device_ui_components_hv("vertical")

    @SafeSlot(dict, dict)
    def on_device_readback_hor(self, msg_content: dict, metadata: dict):
        """Callback for device readback.

        Args:
            msg_content (dict): The message content.
            metadata (dict): The message metadata.
        """
        self._on_device_readback(
            self.device_hor,
            self._device_ui_components_hv("horizontal"),
            msg_content,
            metadata,
            self.position_update_hor.emit,
            self.update_limits_hor,
        )

    @SafeSlot(dict, dict)
    def on_device_readback_ver(self, msg_content: dict, metadata: dict):
        """Callback for device readback.

        Args:
            msg_content (dict): The message content.
            metadata (dict): The message metadata.
        """
        self._on_device_readback(
            self.device_ver,
            self._device_ui_components_hv("vertical"),
            msg_content,
            metadata,
            self.position_update_ver.emit,
            self.update_limits_ver,
        )

    def update_limits_hor(self, limits: tuple):
        """Update limits

        Args:
            limits (tuple): Limits of the positioner
        """
        if limits == self._limits_hor:
            return
        self._limits_hor = limits
        self._update_limits_ui(limits, self.ui.position_indicator_hor, self.setpoint_validator_hor)

    def update_limits_ver(self, limits: tuple):
        """Update limits

        Args:
            limits (tuple): Limits of the positioner
        """
        if limits == self._limits_ver:
            return
        self._limits_ver = limits
        self._update_limits_ui(limits, self.ui.position_indicator_ver, self.setpoint_validator_ver)

    @SafeSlot()
    def on_stop(self):
        self._stop_device(f"{self.device_hor} or {self.device_ver}")

    @SafeProperty(float)
    def step_size_hor(self):
        """Step size for tweak"""
        return self.ui.step_size_hor.value()

    @step_size_hor.setter
    def step_size_hor(self, val: float):
        """Step size for tweak"""
        self.ui.step_size_hor.setValue(val)

    @SafeProperty(float)
    def step_size_ver(self):
        """Step size for tweak"""
        return self.ui.step_size_ver.value()

    @step_size_ver.setter
    def step_size_ver(self, val: float):
        """Step size for tweak"""
        self.ui.step_size_ver.setValue(val)

    @SafeSlot()
    def on_tweak_inc_hor(self):
        """Tweak device a up"""
        self.dev[self.device_hor].move(self.step_size_hor / 10, relative=True)

    @SafeSlot()
    def on_tweak_dec_hor(self):
        """Tweak device a down"""
        self.dev[self.device_hor].move(-self.step_size_hor / 10, relative=True)

    @SafeSlot()
    def on_step_inc_hor(self):
        """Tweak device a up"""
        self.dev[self.device_hor].move(self.step_size_hor, relative=True)

    @SafeSlot()
    def on_step_dec_hor(self):
        """Tweak device a down"""
        self.dev[self.device_hor].move(-self.step_size_hor, relative=True)

    @SafeSlot()
    def on_tweak_inc_ver(self):
        """Tweak device a up"""
        self.dev[self.device_ver].move(self.step_size_ver / 10, relative=True)

    @SafeSlot()
    def on_tweak_dec_ver(self):
        """Tweak device b down"""
        self.dev[self.device_ver].move(-self.step_size_ver / 10, relative=True)

    @SafeSlot()
    def on_step_inc_ver(self):
        """Tweak device b up"""
        self.dev[self.device_ver].move(self.step_size_ver, relative=True)

    @SafeSlot()
    def on_step_dec_ver(self):
        """Tweak device a down"""
        self.dev[self.device_ver].move(-self.step_size_ver, relative=True)

    @SafeSlot()
    def on_setpoint_change_hor(self):
        """Change the setpoint for device a"""
        self.ui.setpoint_hor.clearFocus()
        setpoint = self.ui.setpoint_hor.text()
        self.dev[self.device_hor].move(float(setpoint), relative=False)

    @SafeSlot()
    def on_setpoint_change_ver(self):
        """Change the setpoint for device b"""
        self.ui.setpoint_ver.clearFocus()
        setpoint = self.ui.setpoint_ver.text()
        self.dev[self.device_ver].move(float(setpoint), relative=False)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication  # pylint: disable=ungrouped-imports

    app = QApplication(sys.argv)
    set_theme("dark")
    widget = PositionerBox2D()

    widget.show()
    sys.exit(app.exec_())



"""Utilities for filtering and formatting in the LogPanel"""

from __future__ import annotations

import re
from collections import deque
from typing import Callable, Iterator

from bec_lib.logger import LogLevel
from bec_lib.messages import LogMessage
from qtpy.QtCore import QDateTime

LinesHtmlFormatter = Callable[[deque[LogMessage]], Iterator[str]]
LineFormatter = Callable[[LogMessage], str]
LineFilter = Callable[[LogMessage], bool] | None

ANSI_ESCAPE_REGEX = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def replace_escapes(s: str):
    s = ANSI_ESCAPE_REGEX.sub("", s)
    return s.replace(" ", "&nbsp;").replace("\n", "<br />").replace("\t", "    ")


def level_filter(msg: LogMessage, thresh: int):
    return LogLevel[msg.content["log_type"].upper()].value >= thresh


def noop_format(line: LogMessage):
    _textline = line.log_msg if isinstance(line.log_msg, str) else line.log_msg["text"]
    return replace_escapes(_textline.strip()) + "<br />"


def simple_color_format(line: LogMessage, colors: dict[LogLevel, str]):
    color = colors.get(LogLevel[line.content["log_type"].upper()]) or colors[LogLevel.INFO]
    return f'<font color="{color}">{noop_format(line)}</font>'


def create_formatter(line_format: LineFormatter, line_filter: LineFilter) -> LinesHtmlFormatter:
    def _formatter(data: deque[LogMessage]):
        if line_filter is not None:
            return (line_format(line) for line in data if line_filter(line))
        else:
            return (line_format(line) for line in data)

    return _formatter


def log_txt(line):
    return line.log_msg if isinstance(line.log_msg, str) else line.log_msg["text"]


def log_time(line):
    return QDateTime.fromMSecsSinceEpoch(int(line.log_msg["record"]["time"]["timestamp"] * 1000))


def log_svc(line):
    return line.log_msg["service_name"]
